"""
Simple iterative Incompressible SPH
See https://arxiv.org/abs/1908.01762 for details
"""
import numpy
from numpy import sqrt
from compyle.api import declare
from pysph.sph.scheme import Scheme, add_bool_argument
from pysph.base.utils import get_particle_array
from pysph.sph.integrator import Integrator
from pysph.sph.integrator_step import IntegratorStep
from pysph.sph.equation import Equation, Group, MultiStageEquations


def get_particle_array_sisph(constants=None, **props):
    sisph_props = [
        'u0', 'v0', 'w0', 'x0', 'y0', 'z0', 'rho0', 'diag', 'odiag',
        'pk', 'rhs', 'pdiff', 'wg', 'vf', 'vg', 'ug', 'wij', 'wf', 'uf',
        'V', 'au', 'av', 'aw', 'dt_force', 'dt_cfl', 'vmag',
        'auhat', 'avhat', 'awhat', 'p0', 'uhat', 'vhat', 'what',
        'uhat0', 'vhat0', 'what0', 'pabs'
    ]

    pa = get_particle_array(
        additional_props=sisph_props, constants=constants, **props
    )

    pa.add_constant('iters', [0.0])
    pa.add_constant('pmax', [0.0])
    pa.add_output_arrays(['p', 'V', 'vmag', 'p0'])
    return pa


class SISPHIntegrator(Integrator):
    def one_timestep(self, t, dt):
        self.initialize()

        self.compute_accelerations(0)

        self.stage1()

        self.update_domain()

        self.do_post_stage(0.5*dt, 1)

        self.compute_accelerations(1, update_nnps=False)

        self.stage2()

        self.update_domain()

        self.do_post_stage(dt, 2)

    def initial_acceleration(self, t, dt):
        pass


class SISPHStep(IntegratorStep):
    def initialize(self, d_idx, d_x, d_y, d_z, d_x0, d_y0, d_z0, d_u, d_v,
                   d_w, d_u0, d_v0, d_w0, dt):
        d_x0[d_idx] = d_x[d_idx]
        d_y0[d_idx] = d_y[d_idx]
        d_z0[d_idx] = d_z[d_idx]

        d_u0[d_idx] = d_u[d_idx]
        d_v0[d_idx] = d_v[d_idx]
        d_w0[d_idx] = d_w[d_idx]

    def stage1(self, d_idx, d_u, d_v, d_w, d_au, d_av, d_aw, dt):
        d_u[d_idx] += dt*d_au[d_idx]
        d_v[d_idx] += dt*d_av[d_idx]
        d_w[d_idx] += dt*d_aw[d_idx]

    def stage2(self, d_idx, d_x, d_y, d_z, d_u, d_v, d_w, d_u0, d_v0, d_w0,
               d_x0, d_y0, d_z0, d_au, d_av, d_aw, d_vmag, d_dt_cfl,
               d_dt_force, dt):
        d_u[d_idx] += dt*d_au[d_idx]
        d_v[d_idx] += dt*d_av[d_idx]
        d_w[d_idx] += dt*d_aw[d_idx]

        d_x[d_idx] = d_x0[d_idx] + 0.5*dt * (d_u[d_idx] + d_u0[d_idx])
        d_y[d_idx] = d_y0[d_idx] + 0.5*dt * (d_v[d_idx] + d_v0[d_idx])
        d_z[d_idx] = d_z0[d_idx] + 0.5*dt * (d_w[d_idx] + d_w0[d_idx])

        d_vmag[d_idx] = sqrt(d_u[d_idx]*d_u[d_idx] + d_v[d_idx]*d_v[d_idx] +
                             d_w[d_idx]*d_w[d_idx])

        d_dt_cfl[d_idx] = 2.0*d_vmag[d_idx]

        au = (d_u[d_idx] - d_u0[d_idx])/dt
        av = (d_v[d_idx] - d_v0[d_idx])/dt
        aw = (d_w[d_idx] - d_w0[d_idx])/dt

        d_dt_force[d_idx] = 4.0*(au*au + av*av + aw*aw)


class SISPHGTVFStep(IntegratorStep):
    def initialize(self, d_idx, d_x, d_y, d_z, d_x0, d_y0, d_z0, d_u, d_v,
                   d_w, d_u0, d_v0, d_w0, d_uhat, d_vhat, d_what, d_uhat0,
                   d_vhat0, d_what0):
        d_x0[d_idx] = d_x[d_idx]
        d_y0[d_idx] = d_y[d_idx]
        d_z0[d_idx] = d_z[d_idx]

        d_u0[d_idx] = d_u[d_idx]
        d_v0[d_idx] = d_v[d_idx]
        d_w0[d_idx] = d_w[d_idx]

        d_uhat0[d_idx] = d_uhat[d_idx]
        d_vhat0[d_idx] = d_vhat[d_idx]
        d_what0[d_idx] = d_what[d_idx]

    def stage1(self, d_idx, d_u, d_v, d_w, d_au, d_av, d_aw, dt):
        d_u[d_idx] += dt*d_au[d_idx]
        d_v[d_idx] += dt*d_av[d_idx]
        d_w[d_idx] += dt*d_aw[d_idx]

    def stage2(self, d_idx, d_x, d_y, d_z, d_u, d_v, d_w, d_x0, d_y0, d_z0,
               d_au, d_av, d_aw, d_uhat, d_vhat, d_what, d_auhat, d_avhat,
               d_awhat, d_uhat0, d_vhat0, d_what0, d_vmag, d_dt_cfl, dt,
               d_u0, d_v0, d_w0, d_dt_force):
        d_u[d_idx] += dt*d_au[d_idx]
        d_v[d_idx] += dt*d_av[d_idx]
        d_w[d_idx] += dt*d_aw[d_idx]

        d_vmag[d_idx] = sqrt(d_u[d_idx]*d_u[d_idx] + d_v[d_idx]*d_v[d_idx] +
                             d_w[d_idx]*d_w[d_idx])
        d_dt_cfl[d_idx] = 2.0*d_vmag[d_idx]

        d_uhat[d_idx] = d_u[d_idx] + dt*d_auhat[d_idx]
        d_vhat[d_idx] = d_v[d_idx] + dt*d_avhat[d_idx]
        d_what[d_idx] = d_w[d_idx] + dt*d_awhat[d_idx]

        d_x[d_idx] = d_x0[d_idx] + 0.5*dt * (d_uhat[d_idx] + d_uhat0[d_idx])
        d_y[d_idx] = d_y0[d_idx] + 0.5*dt * (d_vhat[d_idx] + d_vhat0[d_idx])
        d_z[d_idx] = d_z0[d_idx] + 0.5*dt * (d_what[d_idx] + d_what0[d_idx])

        au = (d_u[d_idx] - d_u0[d_idx])/dt
        av = (d_v[d_idx] - d_v0[d_idx])/dt
        aw = (d_w[d_idx] - d_w0[d_idx])/dt

        d_dt_force[d_idx] = 4*(au*au + av*av + aw*aw)


class MomentumEquationBodyForce(Equation):
    def __init__(self, dest, sources, gx=0.0, gy=0.0, gz=0.0):
        self.gx = gx
        self.gy = gy
        self.gz = gz
        super(MomentumEquationBodyForce, self).__init__(dest, sources)

    def initialize(self, d_idx, d_au, d_av, d_aw):
        d_au[d_idx] = 0.0
        d_av[d_idx] = 0.0
        d_aw[d_idx] = 0.0

    def post_loop(self, d_idx, d_au, d_av, d_aw):
        d_au[d_idx] += self.gx
        d_av[d_idx] += self.gy
        d_aw[d_idx] += self.gz


class VelocityDivergence(Equation):
    def initialize(self, d_idx, d_rhs, d_pk, d_p):
        d_rhs[d_idx] = 0.0
        d_pk[d_idx] = d_p[d_idx]

    def loop(self, d_idx, s_idx, s_m, s_rho, d_rhs, dt, VIJ, DWIJ):
        Vj = s_m[s_idx] / s_rho[s_idx]
        vdotdwij = VIJ[0]*DWIJ[0] + VIJ[1]*DWIJ[1] + VIJ[2]*DWIJ[2]
        d_rhs[d_idx] += -Vj * vdotdwij / dt


class VelocityDivergenceSolid(Equation):
    def loop(self, d_idx, s_idx, s_m, s_rho, d_rhs, dt, d_u, d_v, d_w, s_ug,
             s_vg, s_wg, DWIJ):
        Vj = s_m[s_idx] / s_rho[s_idx]
        uij = d_u[d_idx] - s_ug[s_idx]
        vij = d_v[d_idx] - s_vg[s_idx]
        wij = d_w[d_idx] - s_wg[s_idx]
        vdotdwij = uij*DWIJ[0] + vij*DWIJ[1] + wij*DWIJ[2]
        d_rhs[d_idx] += -Vj * vdotdwij / dt


class DensityInvariance(Equation):
    def __init__(self, dest, sources, rho0):
        self.rho0 = rho0
        super(DensityInvariance, self).__init__(dest, sources)

    def post_loop(self, d_idx, d_rho, d_rhs, dt):
        rho0 = self.rho0
        d_rhs[d_idx] = (rho0 - d_rho[d_idx]) / (dt*dt*rho0)


class PressureCoeffMatrixIterative(Equation):
    def initialize(self, d_idx, d_diag, d_odiag):
        d_diag[d_idx] = 0.0
        d_odiag[d_idx] = 0.0

    def loop(self, d_idx, s_idx, s_m, d_rho, s_rho, d_diag, d_odiag, s_pk, XIJ,
             DWIJ, R2IJ, EPS):
        rhoij = (s_rho[s_idx] + d_rho[d_idx])
        rhoij2_1 = 1.0/(d_rho[d_idx]*rhoij)

        xdotdwij = XIJ[0]*DWIJ[0] + XIJ[1]*DWIJ[1] + XIJ[2]*DWIJ[2]

        fac = 4.0 * s_m[s_idx] * rhoij2_1 * xdotdwij / (R2IJ + EPS)

        d_diag[d_idx] += fac
        d_odiag[d_idx] += -fac * s_pk[s_idx]


class PPESolve(Equation):
    def __init__(self, dest, sources, rho0, rho_cutoff=0.8, omega=0.5,
                 tolerance=0.05, max_iterations=1000):
        self.rho0 = rho0
        self.rho_cutoff = rho_cutoff
        self.conv = 0.0
        self.omega = omega
        self.tolerance = tolerance
        self.count = 0.0
        self.max_iterations = max_iterations
        super(PPESolve, self).__init__(dest, sources)

    def post_loop(self, d_idx, d_p, d_pk, d_rhs, d_odiag, d_diag, d_pdiff,
                  d_rho, d_m, d_pabs, d_pmax):
        omega = self.omega
        rho = d_rho[d_idx] / self.rho0
        diag = d_diag[d_idx]
        if abs(diag) < 1e-12 or rho < self.rho_cutoff:
            p = 0.0
        else:
            pnew = (d_rhs[d_idx] - d_odiag[d_idx]) / diag
            p = omega * pnew + (1.0 - omega) * d_pk[d_idx]

        d_pdiff[d_idx] = abs(p - d_pk[d_idx])
        d_pabs[d_idx] = abs(p)
        d_p[d_idx] = p
        d_pk[d_idx] = p
        d_pmax[0] = max(abs(d_pmax[0]), d_p[d_idx])

    def reduce(self, dst, t, dt):
        self.count += 1
        dst.iters[0] = self.count
        if dst.gpu is not None:
            from compyle.array import sum
            n = dst.gpu.p.length
            pdiff = sum(dst.gpu.pdiff, dst.backend) / n
            pmean = sum(dst.gpu.pabs, dst.backend) / n
            conv = pdiff/pmean
            if pmean < 1.0:
                conv = pdiff
            self.conv = 1 if conv < self.tolerance else -1
        else:
            pdiff = dst.pdiff.mean()
            pmean = numpy.abs(dst.p).mean()
            conv = pdiff/pmean
            if pmean < 1.0:
                conv = pdiff
            self.conv = 1 if conv < self.tolerance else -1

    def converged(self):
        if self.conv == 1 and self.count < self.max_iterations:
            self.count = 0
        if self.count > self.max_iterations:
            self.count = 0
            print("Max iterations exceeded")
        return self.conv


class UpdateGhostPressure(Equation):
    def initialize(self, d_idx, d_tag, d_gid, d_p, d_pk):
        idx = declare('int')
        if d_tag[d_idx] == 2:
            idx = d_gid[d_idx]
            d_pk[d_idx] = d_pk[idx]
            d_p[d_idx] = d_p[idx]


class MomentumEquationPressureGradient(Equation):
    def initialize(self, d_idx, d_au, d_av, d_aw):
        d_au[d_idx] = 0.0
        d_av[d_idx] = 0.0
        d_aw[d_idx] = 0.0

    def loop(self, d_idx, s_idx, s_m, d_p, s_p, d_rho, s_rho, d_au,
             d_av, d_aw, DWIJ):
        Vj = s_m[s_idx] / s_rho[s_idx]
        pji = (s_p[s_idx] - d_p[d_idx])
        fac = -Vj * pji / d_rho[d_idx]

        d_au[d_idx] += fac * DWIJ[0]
        d_av[d_idx] += fac * DWIJ[1]
        d_aw[d_idx] += fac * DWIJ[2]


class MomentumEquationPressureGradientSymmetric(Equation):
    def initialize(self, d_idx, d_au, d_av, d_aw):
        d_au[d_idx] = 0.0
        d_av[d_idx] = 0.0
        d_aw[d_idx] = 0.0

    def loop(self, d_idx, s_idx, s_m, d_p, s_p, d_rho, s_rho, d_au, d_av, d_aw,
             DWIJ):
        rhoi2 = d_rho[d_idx]*d_rho[d_idx]
        rhoj2 = s_rho[s_idx]*s_rho[s_idx]
        pij = d_p[d_idx]/rhoi2 + s_p[s_idx]/rhoj2
        fac = -s_m[s_idx] * pij

        d_au[d_idx] += fac * DWIJ[0]
        d_av[d_idx] += fac * DWIJ[1]
        d_aw[d_idx] += fac * DWIJ[2]


class EvaluateNumberDensity(Equation):
    def initialize(self, d_idx, d_wij):
        d_wij[d_idx] = 0.0

    def loop(self, d_idx, d_wij, WIJ):
        d_wij[d_idx] += WIJ


class VolumeSummationBand(Equation):
    def initialize(self, d_idx, d_rhoband):
        d_rhoband[d_idx] = 0.0

    def loop(self, d_idx, d_rhoband, d_m, WIJ):
        d_rhoband[d_idx] += WIJ * d_m[d_idx]


class SetPressureSolid(Equation):
    def __init__(self, dest, sources, gx=0.0, gy=0.0, gz=0.0,
                 hg_correction=True):
        self.gx = gx
        self.gy = gy
        self.gz = gz
        self.hg_correction = hg_correction
        super(SetPressureSolid, self).__init__(dest, sources)

    def initialize(self, d_idx, d_p):
        d_p[d_idx] = 0.0

    def loop(self, d_idx, s_idx, d_p, s_p, s_rho,
             d_au, d_av, d_aw, WIJ, XIJ):

        # numerator of Eq. (27) ax, ay and az are the prescribed wall
        # accelerations which must be defined for the wall boundary
        # particle
        gdotxij = (self.gx - d_au[d_idx])*XIJ[0] + \
            (self.gy - d_av[d_idx])*XIJ[1] + \
            (self.gz - d_aw[d_idx])*XIJ[2]

        d_p[d_idx] += s_p[s_idx]*WIJ + s_rho[s_idx]*gdotxij*WIJ

    def post_loop(self, d_idx, d_wij, d_p, d_rho, d_pk):
        # extrapolated pressure at the ghost particle
        if d_wij[d_idx] > 1e-14:
            d_p[d_idx] /= d_wij[d_idx]
        if self.hg_correction:
            d_p[d_idx] = max(0.0, d_p[d_idx])
        d_pk[d_idx] = d_p[d_idx]


class GTVFAcceleration(Equation):
    def __init__(self, dest, sources, pref, internal_flow=False,
                 use_pref=False):
        self.pref = pref
        assert self.pref is not None, "pref should not be None"
        self.internal = internal_flow
        self.hij_fac = 1 if self.internal else 0.5
        self.use_pref = use_pref
        super(GTVFAcceleration, self).__init__(dest, sources)

    def initialize(self, d_idx, d_auhat, d_avhat, d_awhat, d_p0, d_p, d_pmax):
        d_auhat[d_idx] = 0.0
        d_avhat[d_idx] = 0.0
        d_awhat[d_idx] = 0.0

        if self.internal:
            if self.use_pref:
                d_p0[d_idx] = self.pref
            else:
                pref = 2*d_pmax[0]
                d_p0[d_idx] = pref
        else:
            d_p0[d_idx] = min(10*abs(d_p[d_idx]), self.pref)

    def loop(self, d_p0, s_m, s_idx, d_rho, d_idx, d_auhat, d_avhat,
             d_awhat, XIJ, RIJ, SPH_KERNEL,
             HIJ):
        rhoi2 = d_rho[d_idx]*d_rho[d_idx]
        tmp = -d_p0[d_idx] * s_m[s_idx]/rhoi2

        dwijhat = declare('matrix(3)')
        SPH_KERNEL.gradient(XIJ, RIJ, self.hij_fac*HIJ, dwijhat)

        d_auhat[d_idx] += tmp * dwijhat[0]
        d_avhat[d_idx] += tmp * dwijhat[1]
        d_awhat[d_idx] += tmp * dwijhat[2]


class SmoothedVelocity(Equation):
    def initialize(self, d_ax, d_ay, d_az, d_idx):
        d_ax[d_idx] = 0.0
        d_ay[d_idx] = 0.0
        d_az[d_idx] = 0.0

    def loop(self, d_ax, d_ay, d_az, d_idx, s_uhat, s_vhat, s_what, s_idx, s_m,
             s_rho, WIJ):
        fac = s_m[s_idx]*WIJ / s_rho[s_idx]
        d_ax[d_idx] += fac*s_uhat[s_idx]
        d_ay[d_idx] += fac*s_vhat[s_idx]
        d_az[d_idx] += fac*s_what[s_idx]


class SolidWallNoSlipBC(Equation):
    def __init__(self, dest, sources, nu):
        self.nu = nu
        super(SolidWallNoSlipBC, self).__init__(dest, sources)

    def loop(self, d_idx, s_idx, d_m, d_rho, s_rho, s_m,
             d_u, d_v, d_w,
             d_au, d_av, d_aw,
             s_ug, s_vg, s_wg,
             DWIJ, R2IJ, EPS, XIJ):
        mj = s_m[s_idx]
        rhoi = d_rho[d_idx]
        rhoj = s_rho[s_idx]
        rhoij1 = 1.0/(rhoi + rhoj)

        Fij = XIJ[0]*DWIJ[0] + XIJ[1]*DWIJ[1] + XIJ[2]*DWIJ[2]

        tmp = mj * 4 * self.nu * rhoij1 * Fij/(R2IJ + EPS)

        d_au[d_idx] += tmp * (d_u[d_idx] - s_ug[s_idx])
        d_av[d_idx] += tmp * (d_v[d_idx] - s_vg[s_idx])
        d_aw[d_idx] += tmp * (d_w[d_idx] - s_wg[s_idx])


class SummationDensity(Equation):
    def initialize(self, d_idx, d_rho):
        d_rho[d_idx] = 0.0

    def loop(self, d_idx, s_idx, d_rho, s_m, WIJ):
        d_rho[d_idx] += s_m[s_idx]*WIJ


class SISPHScheme(Scheme):
    def __init__(self, fluids, solids, dim, nu, rho0, c0, alpha=0.0, beta=0.0,
                 gx=0.0, gy=0.0, gz=0.0, tolerance=0.05,
                 omega=0.5, hg_correction=False, has_ghosts=False,
                 pref=None, gtvf=False, symmetric=False, rho_cutoff=0.8,
                 max_iterations=1000, internal_flow=False,
                 use_pref=False):
        self.fluids = fluids
        self.solids = solids
        self.solver = None
        self.dim = dim
        self.nu = nu
        self.gx = gx
        self.gy = gy
        self.gz = gz
        self.c0 = c0
        self.alpha = alpha
        self.beta = beta
        self.rho0 = rho0
        self.rho_cutoff = rho_cutoff
        self.tolerance = tolerance
        self.omega = omega
        self.hg_correction = hg_correction
        self.has_ghosts = has_ghosts
        self.pref = pref
        self.gtvf = gtvf
        self.symmetric = symmetric
        self.max_iterations = max_iterations
        self.internal_flow = internal_flow
        self.use_pref = use_pref

    def add_user_options(self, group):
        group.add_argument(
            "--tol", action="store", dest="tolerance",
            type=float,
            help="Tolerance for convergence."
        )
        group.add_argument(
            "--omega", action="store", dest="omega",
            type=float,
            help="Omega for convergence."
        )
        group.add_argument(
            '--alpha', action='store', type=float, dest='alpha',
            default=None,
            help='Artificial viscosity.'
        )
        add_bool_argument(
            group, 'gtvf', dest='gtvf', default=None,
            help='Use GTVF.'
        )
        add_bool_argument(
            group, 'symmetric', dest='symmetric', default=None,
            help='Use symmetric form of pressure gradient.'
        )
        add_bool_argument(
            group, 'internal', dest='internal_flow', default=None,
            help='If the simulation is internal or external.'
        )

    def consume_user_options(self, options):
        _vars = ['tolerance', 'omega', 'alpha', 'gtvf', 'symmetric',
                 'internal_flow']
        data = dict((var, self._smart_getattr(options, var))
                    for var in _vars)
        self.configure(**data)

    def configure_solver(self, kernel=None, integrator_cls=None,
                         extra_steppers=None, **kw):
        import pysph.base.kernels as kern
        if kernel is None:
            kernel = kern.QuinticSpline(dim=self.dim)
        steppers = {}
        if extra_steppers is not None:
            steppers.update(extra_steppers)

        step_cls = SISPHStep
        if self.gtvf:
            step_cls = SISPHGTVFStep

        for fluid in self.fluids:
            if fluid not in steppers:
                steppers[fluid] = step_cls()

        if integrator_cls is not None:
            cls = integrator_cls
        else:
            cls = SISPHIntegrator

        integrator = cls(**steppers)

        from pysph.solver.solver import Solver
        self.solver = Solver(
            dim=self.dim, integrator=integrator, kernel=kernel, **kw
        )

    def _get_velocity_bc(self):
        from pysph.sph.isph.wall_normal import SetWallVelocityNew

        eqs = [SetWallVelocityNew(dest=s, sources=self.fluids)
               for s in self.solids]
        return Group(equations=eqs)

    def _get_pressure_bc(self):
        eqs = []
        all_solids = self.solids
        for solid in all_solids:
            eqs.append(
                EvaluateNumberDensity(
                    dest=solid, sources=self.fluids
                )
            )
            eqs.append(
                SetPressureSolid(
                    dest=solid, sources=self.fluids,
                    gx=self.gx, gy=self.gy, gz=self.gz,
                    hg_correction=self.hg_correction
                )
            )

        return Group(equations=eqs) if eqs else None

    def _get_normals(self, pa):
        from pysph.tools.sph_evaluator import SPHEvaluator
        from pysph.sph.isph.wall_normal import ComputeNormals, SmoothNormals

        pa.add_property('normal', stride=3)
        pa.add_property('normal_tmp', stride=3)

        name = pa.name

        seval = SPHEvaluator(
            arrays=[pa], equations=[
                Group(equations=[
                    ComputeNormals(dest=name, sources=[name])
                ]),
                Group(equations=[
                    SmoothNormals(dest=name, sources=[name])
                ]),
            ],
            dim=self.dim
        )
        seval.evaluate()

    def _get_viscous_eqns(self):
        from pysph.sph.wc.transport_velocity import (
            MomentumEquationArtificialViscosity)
        from pysph.sph.wc.viscosity import LaminarViscosity
        from pysph.sph.wc.gtvf import MomentumEquationArtificialStress

        all = self.fluids + self.solids

        eq, stg = [], []
        for fluid in self.fluids:
            eq.append(SummationDensity(dest=fluid, sources=all))
        stg.append(Group(equations=eq, real=False))

        eq = []
        for fluid in self.fluids:
            if self.nu > 0.0:
                eq.append(
                    LaminarViscosity(fluid, sources=self.fluids,
                                     nu=self.nu)
                )
            if self.alpha > 0.0:
                eq.append(
                    MomentumEquationArtificialViscosity(
                        dest=fluid, sources=self.fluids, c0=self.c0,
                        alpha=self.alpha
                    )
                )
            eq.append(
                MomentumEquationBodyForce(
                    fluid, sources=None, gx=self.gx, gy=self.gy,
                    gz=self.gz)
            )
            if self.gtvf:
                eq.append(
                    MomentumEquationArtificialStress(
                        dest=fluid, sources=self.fluids, dim=self.dim)
                )
            if self.solids and self.nu > 0.0:
                eq.append(
                    SolidWallNoSlipBC(
                        dest=fluid, sources=self.solids, nu=self.nu
                    )
                )
        stg.append(Group(equations=eq))
        return stg

    def _get_ppe(self):
        from pysph.sph.wc.transport_velocity import VolumeSummation

        all = self.fluids + self.solids
        all_solids = self.solids

        eq, stg = [], []

        for fluid in self.fluids:
            eq.append(SummationDensity(dest=fluid, sources=all))
        stg.append(Group(equations=eq, real=False))

        eq2 = []
        for fluid in self.fluids:
            eq2.append(VolumeSummation(dest=fluid, sources=all))
            eq2.append(
                VelocityDivergence(
                    dest=fluid,
                    sources=self.fluids
                ))
            if self.solids:
                eq2.append(
                    VelocityDivergenceSolid(fluid, sources=self.solids)
                )
        stg.append(Group(equations=eq2))

        solver_eqns = []
        if self.has_ghosts:
            ghost_eqns = Group(
                equations=[UpdateGhostPressure(dest=fluid, sources=None)
                           for fluid in self.fluids],
                real=False
            )
            solver_eqns = [ghost_eqns]

        if all_solids:
            g3 = self._get_pressure_bc()
            solver_eqns.append(g3)

        eq3 = []
        for fluid in self.fluids:
            if not fluid == 'outlet':
                eq3.append(
                    PressureCoeffMatrixIterative(dest=fluid, sources=all)
                )
                eq3.append(
                    PPESolve(
                        dest=fluid, sources=all, rho0=self.rho0,
                        rho_cutoff=self.rho_cutoff, tolerance=self.tolerance,
                        omega=self.omega, max_iterations=self.max_iterations
                    )
                )
        eq3 = Group(equations=eq3)

        solver_eqns.append(eq3)

        stg.append(
            Group(
                equations=solver_eqns, iterate=True,
                max_iterations=self.max_iterations, min_iterations=2
            )
        )

        if self.has_ghosts:
            ghost_eqns = Group(
                equations=[UpdateGhostPressure(dest=fluid, sources=None)
                           for fluid in self.fluids],
                real=False
            )
            stg.append(ghost_eqns)
        return stg

    def get_equations(self):
        all = self.fluids + self.solids
        all_solids = self.solids

        stg1 = []
        if all_solids:
            g0 = self._get_velocity_bc()
            stg1.append(g0)

        stg1.extend(self._get_viscous_eqns())

        stg2 = []
        if all_solids:
            g0 = self._get_velocity_bc()
            stg2.append(g0)

        stg2.extend(self._get_ppe())

        if all_solids:
            g3 = self._get_pressure_bc()
            stg2.append(g3)

        if all_solids:
            g0 = self._get_velocity_bc()
            stg2.append(g0)

        eq4 = []
        for fluid in self.fluids:
            if self.symmetric:
                eq4.append(
                    MomentumEquationPressureGradientSymmetric(fluid, all)
                )
            else:
                eq4.append(
                    MomentumEquationPressureGradient(fluid, sources=all)
                )
            if self.gtvf:
                eq4.append(
                    GTVFAcceleration(dest=fluid, sources=all, pref=self.pref,
                                     internal_flow=self.internal_flow,
                                     use_pref=self.use_pref)
                )
        stg2.append(Group(equations=eq4))
        return MultiStageEquations([stg1, stg2])

    def setup_properties(self, particles, clean=True):
        particle_arrays = dict([(p.name, p) for p in particles])
        dummy = get_particle_array_sisph(
            name='junk', gid=particle_arrays['fluid'].gid
        )
        props = list(dummy.properties.keys())
        props += [dict(name=x, stride=v) for x, v in dummy.stride.items()]
        constants = [dict(name=x, data=v) for x, v in dummy.constants.items()]
        output_props = dummy.output_property_arrays
        for fluid in self.fluids:
            pa = particle_arrays[fluid]
            self._ensure_properties(pa, props, clean)
            pa.set_output_arrays(output_props)
            for const in constants:
                pa.add_constant(**const)

        solid_props = ['wij', 'ug', 'vg', 'wg', 'uf', 'vf', 'wf', 'pk', 'V']
        all_solids = self.solids
        for solid in all_solids:
            pa = particle_arrays[solid]
            for prop in solid_props:
                pa.add_property(prop)
            self._get_normals(pa)
            pa.add_output_arrays(['p', 'ug', 'vg', 'wg', 'normal'])



from math import sqrt
from pysph.sph.equation import Equation
from compyle.api import declare
from pysph.sph.wc.linalg import gj_solve, augmented_matrix


class ShepardFilter(Equation):
    r"""**Shepard Filter density reinitialization**
    This is a zeroth order density reinitialization

    .. math::
            \tilde{W_{ab}} = \frac{W_{ab}}{\sum_{b} W_{ab}\frac{m_{b}}
            {\rho_{b}}}

    .. math::
            \rho_{a} = \sum_{b} \m_{b}\tilde{W_{ab}}

    References
    ----------
    .. [Panizzo, 2004] Panizzo, Physical and Numerical Modelling of
        Subaerial Landslide Generated Waves, PhD thesis.
    """

    def initialize(self, d_idx, d_rho, d_rhotmp):
        d_rhotmp[d_idx] = d_rho[d_idx]

    def loop_all(self, d_idx, d_rho, d_x, d_y, d_z, s_m, s_rhotmp, s_x, s_y,
                 s_z, d_h, s_h, SPH_KERNEL, NBRS, N_NBRS):
        i, s_idx = declare('int', 2)
        xij = declare('matrix(3)')
        tmp_w = 0.0
        x = d_x[d_idx]
        y = d_y[d_idx]
        z = d_z[d_idx]
        d_rho[d_idx] = 0.0
        for i in range(N_NBRS):
            s_idx = NBRS[i]
            xij[0] = x - s_x[s_idx]
            xij[1] = y - s_y[s_idx]
            xij[2] = z - s_z[s_idx]
            rij = sqrt(xij[0] * xij[0] + xij[1] * xij[1] + xij[2] * xij[2])
            hij = (d_h[d_idx] + s_h[s_idx]) * 0.5
            wij = SPH_KERNEL.kernel(xij, rij, hij)
            tmp_w += wij * s_m[s_idx] / s_rhotmp[s_idx]
            d_rho[d_idx] += wij * s_m[s_idx]
        d_rho[d_idx] /= tmp_w


class MLSFirstOrder2D(Equation):
    r"""**Moving Least Squares density reinitialization**
    This is a first order density reinitialization

    .. math::
            W_{ab}^{MLS} = \beta\left(\mathbf{r_{a}}\right)\cdot\left(
            \mathbf{r}_a - \mathbf{r}_b\right)W_{ab}

    .. math::
            \beta\left(\mathbf{r_{a}}\right) = A^{-1}
            \left[1 0 0\right]^{T}

    where

    .. math::
            A = \sum_{b}W_{ab}\tilde{A}\frac{m_{b}}{\rho_{b}}

    .. math::
            \tilde{A} = pp^{T}

    where

    .. math::
            p = \left[1 x_{a}-x_{b} y_{a}-y_{b}\right]^{T}

    .. math::
            \rho_{a} = \sum_{b} \m_{b}W_{ab}^{MLS}
    References
    ----------
    .. [Dilts, 1999] Dilts, G. A. Moving-Least-Squares-Particle
        Hydrodynamics - I. Consistency and stability,
        Int. J. Numer. Meth. Engng, 1999.
    """

    def _get_helpers_(self):
        return [gj_solve, augmented_matrix]

    def initialize(self, d_idx, d_rho, d_rhotmp):
        d_rhotmp[d_idx] = d_rho[d_idx]

    def loop_all(self, d_idx, d_rho, d_x, d_y, s_x, s_y, d_h, s_h, s_m,
                 s_rhotmp, SPH_KERNEL, NBRS, N_NBRS):
        n, i, j, k, s_idx = declare('int', 5)
        n = 3
        amls = declare('matrix(9)')
        aug_mls = declare('matrix(12)')
        x = d_x[d_idx]
        y = d_y[d_idx]
        xij = declare('matrix(3)')
        for i in range(n):
            for j in range(n):
                amls[n * i + j] = 0.0
        for k in range(N_NBRS):
            s_idx = NBRS[k]
            xij[0] = x - s_x[s_idx]
            xij[1] = y - s_y[s_idx]
            xij[2] = 0.
            rij = sqrt(xij[0] * xij[0] + xij[1] * xij[1])
            hij = (d_h[d_idx] + s_h[s_idx]) * 0.5
            wij = SPH_KERNEL.kernel(xij, rij, hij)
            for i in range(n):
                if i == 0:
                    fac1 = 1.0
                else:
                    fac1 = xij[i - 1]
                for j in range(n):
                    if j == 0:
                        fac2 = 1.0
                    else:
                        fac2 = xij[j - 1]
                    amls[n * i + j] += fac1 * fac2 * \
                        s_m[s_idx] * wij / s_rhotmp[s_idx]
        res = declare('matrix(3)')
        res[0] = 1.0
        res[1] = 0.0
        res[2] = 0.0
        augmented_matrix(amls, res, 3, 1, 3, aug_mls)
        gj_solve(aug_mls, n, 1, res)
        b0 = res[0]
        b1 = res[1]
        b2 = res[2]
        d_rho[d_idx] = 0.0
        for k in range(N_NBRS):
            s_idx = NBRS[k]
            xij[0] = x - s_x[s_idx]
            xij[1] = y - s_y[s_idx]
            xij[2] = 0.
            rij = sqrt(xij[0] * xij[0] + xij[1] * xij[1])
            hij = (d_h[d_idx] + s_h[s_idx]) * 0.5
            wij = SPH_KERNEL.kernel(xij, rij, hij)
            wmls = (b0 + b1 * xij[0] + b2 * xij[1]) * wij
            d_rho[d_idx] += s_m[s_idx] * wmls


class MLSFirstOrder3D(Equation):

    def _get_helpers_(self):
        return [gj_solve, augmented_matrix]

    def initialize(self, d_idx, d_rho, d_rhotmp):
        d_rhotmp[d_idx] = d_rho[d_idx]

    def loop_all(self, d_idx, d_rho, d_x, d_y, d_z, s_x, s_y, s_z, d_h, s_h,
                 s_m, s_rhotmp, SPH_KERNEL, NBRS, N_NBRS):
        n, i, j, k, s_idx = declare('int', 5)
        n = 4
        amls = declare('matrix(16)')
        aug_mls = declare('matrix(20)')
        x = d_x[d_idx]
        y = d_y[d_idx]
        z = d_z[d_idx]
        xij = declare('matrix(4)')
        for i in range(n):
            for j in range(n):
                amls[n * i + j] = 0.0
        for k in range(N_NBRS):
            s_idx = NBRS[k]
            xij[0] = x - s_x[s_idx]
            xij[1] = y - s_y[s_idx]
            xij[2] = z - s_z[s_idx]
            rij = sqrt(xij[0] * xij[0] + xij[1] * xij[1] + xij[2] * xij[2])
            hij = (d_h[d_idx] + s_h[s_idx]) * 0.5
            wij = SPH_KERNEL.kernel(xij, rij, hij)
            for i in range(n):
                if i == 0:
                    fac1 = 1.0
                else:
                    fac1 = xij[i - 1]
                for j in range(n):
                    if j == 0:
                        fac2 = 1.0
                    else:
                        fac2 = xij[j - 1]
                    amls[n * i + j] += fac1 * fac2 * \
                        s_m[s_idx] * wij / s_rhotmp[s_idx]
        res = declare('matrix(4)')
        res[0] = 1.0
        res[1] = 0.0
        res[2] = 0.0
        res[3] = 0.0
        augmented_matrix(amls, res, n, 1, aug_mls)
        gj_solve(aug_mls, n, 1, res)
        b0 = res[0]
        b1 = res[1]
        b2 = res[2]
        b3 = res[3]
        d_rho[d_idx] = 0.0
        for k in range(N_NBRS):
            s_idx = NBRS[k]
            xij[0] = x - s_x[s_idx]
            xij[1] = y - s_y[s_idx]
            xij[2] = z - s_z[s_idx]
            rij = sqrt(xij[0] * xij[0] + xij[1] * xij[1] + xij[2] * xij[2])
            hij = (d_h[d_idx] + s_h[s_idx]) * 0.5
            wij = SPH_KERNEL.kernel(xij, rij, hij)
            wmls = (b0 + b1 * xij[0] + b2 * xij[1] + b3 * xij[2]) * wij
            d_rho[d_idx] += s_m[s_idx] * wmls



# cython: language_level=3, embedsignature=True
# distutils: language=c++

# malloc and friends
from libc.stdlib cimport malloc, free
from libcpp.vector cimport vector

# Cython for compiler directives
cimport cython

IF UNAME_SYSNAME == "Windows":
    cdef inline double fmin(double x, double y) nogil:
        return x if x < y else y
    cdef inline double fmax(double x, double y) nogil:
        return x if x > y else y


#############################################################################
cdef class SpatialHashNNPS(NNPS):

    """Nearest neighbor particle search using Spatial Hashing algorithm

    Uses a hashtable to store particles according to cell it belongs to.

    Ref. http://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.105.6732&rep=rep1&type=pdf
    """

    def __init__(self, int dim, list particles, double radius_scale = 2.0,
            int ghost_layers = 1, domain=None,
            bint fixed_h = False, bint cache = False,
            bint sort_gids = False, long long int table_size = 131072):
        #Initialize base class
        NNPS.__init__(
            self, dim, particles, radius_scale, ghost_layers, domain,
            cache, sort_gids
        )

        self.src_index = 0
        self.dst_index = 0
        self.sort_gids = sort_gids
        self.domain.update()
        self.update()

    def __cinit__(self, int dim, list particles, double radius_scale = 2.0,
            int ghost_layers = 1, domain=None,
            bint fixed_h = False, bint cache = False,
            bint sort_gids = False, long long int table_size = 131072):

        cdef int narrays = len(particles)

        self.table_size = table_size
        self.radius_scale2 = radius_scale*radius_scale

        self.hashtable = <HashTable**> malloc(narrays*sizeof(HashTable*))

        cdef int i
        for i from 0<=i<narrays:
            self.hashtable[i] = new HashTable(table_size)

        self.current_hash = NULL

    def __dealloc__(self):
        cdef int i
        for i from 0<=i<self.narrays:
            del self.hashtable[i]
        free(self.hashtable)


    #### Public protocol ################################################

    cpdef set_context(self, int src_index, int dst_index):
        """Set context for nearest neighbor searches.

        Parameters
        ----------
        src_index: int
            Index in the list of particle arrays to which the neighbors belong

        dst_index: int
            Index in the list of particle arrays to which the query point belongs

        """
        NNPS.set_context(self, src_index, dst_index)
        self.current_hash = self.hashtable[src_index]

        self.dst = <NNPSParticleArrayWrapper> self.pa_wrappers[dst_index]
        self.src = <NNPSParticleArrayWrapper> self.pa_wrappers[src_index]

    cdef void find_nearest_neighbors(self, size_t d_idx, UIntArray nbrs) nogil:
        """Low level, high-performance non-gil method to find neighbors.
        This requires that `set_context()` be called beforehand.  This method
        does not reset the neighbors array before it appends the
        neighbors to it.

        """
        cdef double* dst_x_ptr = self.dst.x.data
        cdef double* dst_y_ptr = self.dst.y.data
        cdef double* dst_z_ptr = self.dst.z.data
        cdef double* dst_h_ptr = self.dst.h.data

        cdef double* src_x_ptr = self.src.x.data
        cdef double* src_y_ptr = self.src.y.data
        cdef double* src_z_ptr = self.src.z.data
        cdef double* src_h_ptr = self.src.h.data

        cdef double x = dst_x_ptr[d_idx]
        cdef double y = dst_y_ptr[d_idx]
        cdef double z = dst_z_ptr[d_idx]
        cdef double h = dst_h_ptr[d_idx]

        cdef unsigned int* s_gid = self.src.gid.data
        cdef int orig_length = nbrs.length

        cdef int c_x, c_y, c_z
        cdef double* xmin = self.xmin.data
        cdef unsigned int i, j, k

        cdef HashEntry* candidate_cell
        cdef vector[unsigned int] *candidates

        find_cell_id_raw(
                x - xmin[0],
                y - xmin[1],
                z - xmin[2],
                self.cell_size,
                &c_x, &c_y, &c_z
                )
        cdef int candidate_size = 0

        cdef int x_boxes[27]
        cdef int y_boxes[27]
        cdef int z_boxes[27]
        cdef int num_boxes = self._neighbor_boxes(c_x, c_y, c_z,
                x_boxes, y_boxes, z_boxes)

        cdef double xij2 = 0
        cdef double hi2 = self.radius_scale2*h*h
        cdef double hj2 = 0

        for i from 0<=i<num_boxes:
            candidate_cell = self.current_hash.get(x_boxes[i], y_boxes[i], z_boxes[i])
            if candidate_cell == NULL:
                continue
            candidates = candidate_cell.get_indices()
            candidate_size = candidates.size()
            for j from 0<=j<candidate_size:
                k = (candidates[0])[j]
                hj2 = self.radius_scale2*src_h_ptr[k]*src_h_ptr[k]
                xij2 = norm2(
                        src_x_ptr[k] - x,
                        src_y_ptr[k] - y,
                        src_z_ptr[k] - z
                        )
                if (xij2 < hi2) or (xij2 < hj2):
                    nbrs.c_append(k)

        if self.sort_gids:
            self._sort_neighbors(
                &nbrs.data[orig_length], nbrs.length - orig_length, s_gid
            )


    #### Private protocol ################################################

    cdef inline void _add_to_hashtable(self, int hash_id, unsigned int pid, double h,
            int i, int j, int k) nogil:
        self.hashtable[hash_id].add(i,j,k,pid,h)

    cdef inline int _neighbor_boxes(self, int i, int j, int k,
            int* x, int* y, int* z) nogil:
        cdef int length = 0
        cdef int p, q, r
        for p from -1<=p<2:
            for q from -1<=q<2:
                for r from -1<=r<2:
                    if i+p>=0 and j+q>=0 and k+r>=0:
                        x[length] = i+p
                        y[length] = j+q
                        z[length] = k+r
                        length += 1
        return length

    cpdef _refresh(self):
        cdef int i
        for i from 0<=i<self.narrays:
            del self.hashtable[i]
            self.hashtable[i] = new HashTable(self.table_size)
        self.current_hash = self.hashtable[self.src_index]

    cpdef _bin(self, int pa_index, UIntArray indices):
        cdef NNPSParticleArrayWrapper pa_wrapper = self.pa_wrappers[pa_index]

        cdef double* src_x_ptr = pa_wrapper.x.data
        cdef double* src_y_ptr = pa_wrapper.y.data
        cdef double* src_z_ptr = pa_wrapper.z.data
        cdef double* src_h_ptr = pa_wrapper.h.data

        cdef int num_indices = indices.length

        cdef double* xmin = self.xmin.data
        cdef int c_x, c_y, c_z
        cdef unsigned int i
        cdef unsigned int idx

        for i from 0<=i<num_indices:
            idx = indices.data[i]
            find_cell_id_raw(
                    src_x_ptr[idx] - xmin[0],
                    src_y_ptr[idx] - xmin[1],
                    src_z_ptr[idx] - xmin[2],
                    self.cell_size,
                    &c_x, &c_y, &c_z
                    )
            self._add_to_hashtable(pa_index, idx, src_h_ptr[idx], c_x, c_y, c_z)


#############################################################################
cdef class ExtendedSpatialHashNNPS(NNPS):

    """Finds nearest neighbors using Extended Spatial Hashing algorithm

    Sub-divides each cell into smaller ones. Useful when particles cluster
    in a cell.

    For approximate Extended Spatial Hash, if the distance between a cell and
    the cell of the query particle is greater than search radius, the entire cell
    is ignored.

    Ref. http://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.105.6732&rep=rep1&type=pdf
    """

    def __init__(self, int dim, list particles, double radius_scale = 2.0,
            int H = 3, int ghost_layers = 1, domain=None, bint fixed_h = False,
            bint cache = False, bint sort_gids = False,
            long long int table_size = 131072, bint approximate = False):
        NNPS.__init__(
            self, dim, particles, radius_scale, ghost_layers, domain,
            cache, sort_gids
        )

        self.H = H
        self.approximate = approximate

        self.src_index = 0
        self.dst_index = 0
        self.sort_gids = sort_gids
        self.domain.update()
        self.update()

    def __cinit__(self, int dim, list particles, double radius_scale = 2.0,
            int H = 3, int ghost_layers = 1, domain=None, bint fixed_h = False,
            bint cache = False, bint sort_gids = False,
            long long int table_size = 131072, bint approximate = False):

        cdef int narrays = len(particles)

        self.table_size = table_size
        self.radius_scale2 = radius_scale*radius_scale

        self.hashtable = <HashTable**> malloc(narrays*sizeof(HashTable*))

        cdef int i
        for i from 0<=i<narrays:
            self.hashtable[i] = new HashTable(table_size)

        self.current_hash = NULL

    def __dealloc__(self):
        cdef int i
        for i from 0<=i<self.narrays:
            del self.hashtable[i]
        free(self.hashtable)


    #### Public protocol ################################################

    cpdef set_context(self, int src_index, int dst_index):
        """Set context for nearest neighbor searches.

        Parameters
        ----------
        src_index: int
            Index in the list of particle arrays to which the neighbors belong

        dst_index: int
            Index in the list of particle arrays to which the query point belongs

        """
        NNPS.set_context(self, src_index, dst_index)
        self.current_hash = self.hashtable[src_index]

        self.dst = <NNPSParticleArrayWrapper> self.pa_wrappers[dst_index]
        self.src = <NNPSParticleArrayWrapper> self.pa_wrappers[src_index]

    cdef void find_nearest_neighbors(self, size_t d_idx, UIntArray nbrs) nogil:
        """Low level, high-performance non-gil method to find neighbors.
        This requires that `set_context()` be called beforehand.  This method
        does not reset the neighbors array before it appends the
        neighbors to it.

        """
        cdef double* dst_x_ptr = self.dst.x.data
        cdef double* dst_y_ptr = self.dst.y.data
        cdef double* dst_z_ptr = self.dst.z.data
        cdef double* dst_h_ptr = self.dst.h.data

        cdef double* src_x_ptr = self.src.x.data
        cdef double* src_y_ptr = self.src.y.data
        cdef double* src_z_ptr = self.src.z.data
        cdef double* src_h_ptr = self.src.h.data

        cdef double x = dst_x_ptr[d_idx]
        cdef double y = dst_y_ptr[d_idx]
        cdef double z = dst_z_ptr[d_idx]
        cdef double h = dst_h_ptr[d_idx]

        cdef unsigned int* s_gid = self.src.gid.data
        cdef int orig_length = nbrs.length

        cdef int c_x, c_y, c_z
        cdef double* xmin = self.xmin.data
        cdef unsigned int i, j, k

        cdef HashEntry* candidate_cell
        cdef vector[unsigned int] *candidates

        find_cell_id_raw(
                x - xmin[0],
                y - xmin[1],
                z - xmin[2],
                self.h_sub,
                &c_x, &c_y, &c_z
                )

        cdef int candidate_size = 0
        cdef int mask_len = (2*self.H+1)*(2*self.H+1)*(2*self.H+1)

        cdef int* x_boxes = <int*> malloc(mask_len*sizeof(int))
        cdef int* y_boxes = <int*> malloc(mask_len*sizeof(int))
        cdef int* z_boxes = <int*> malloc(mask_len*sizeof(int))

        cdef int num_boxes = self._neighbor_boxes(c_x, c_y, c_z,
                x_boxes, y_boxes, z_boxes, h)

        cdef double xij2 = 0
        cdef double hi2 = self.radius_scale2*h*h
        cdef double hj2 = 0

        for i from 0<=i<num_boxes:
            candidate_cell = self.current_hash.get(x_boxes[i], y_boxes[i], z_boxes[i])
            if candidate_cell == NULL:
                continue
            candidates = candidate_cell.get_indices()
            candidate_size = candidates.size()
            for j from 0<=j<candidate_size:
                k = (candidates[0])[j]
                hj2 = self.radius_scale2*src_h_ptr[k]*src_h_ptr[k]
                xij2 = norm2(
                        src_x_ptr[k] - x,
                        src_y_ptr[k] - y,
                        src_z_ptr[k] - z
                        )
                if (xij2 < hi2) or (xij2 < hj2):
                    nbrs.c_append(k)

        free(x_boxes)
        free(y_boxes)
        free(z_boxes)

        if self.sort_gids:
            self._sort_neighbors(
                &nbrs.data[orig_length], nbrs.length - orig_length, s_gid
            )


    #### Private protocol ################################################

    cdef inline void _add_to_hashtable(self, int hash_id, unsigned int pid, double h,
            int i, int j, int k) nogil:
        self.hashtable[hash_id].add(i,j,k,pid,h)

    cdef inline int _h_mask_approx(self, int* x, int* y, int* z) nogil:
        cdef int length = 0
        cdef int s, t, u

        for s from -self.H<=s<=self.H:
            for t from -self.H<=t<=self.H:
                for u from -self.H<=u<=self.H:
                    if norm2(self.h_sub*s, self.h_sub*t, self.h_sub*u) \
                        <= self.cell_size*self.cell_size:
                            x[length] = s
                            y[length] = t
                            z[length] = u
                            length += 1

        return length

    cdef inline int _h_mask_exact(self, int* x, int* y, int* z) nogil:
        cdef int length = 0
        cdef int s, t, u

        for s from -self.H<=s<=self.H:
            for t from -self.H<=t<=self.H:
                for u from -self.H<=u<=self.H:

                    x[length] = s
                    y[length] = t
                    z[length] = u
                    length += 1

        return length

    @cython.cdivision(True)
    cdef int _neighbor_boxes(self, int i, int j, int k,
            int* x, int* y, int* z, double h) nogil:
        cdef int length = 0
        cdef int p

        cdef int x_temp, y_temp, z_temp

        cdef HashEntry* cell
        cdef double h_local
        cdef int H

        cdef int mask_len = (2*self.H+1)*(2*self.H+1)*(2*self.H+1)

        cdef int* x_mask = <int*> malloc(mask_len*sizeof(int))
        cdef int* y_mask = <int*> malloc(mask_len*sizeof(int))
        cdef int* z_mask = <int*> malloc(mask_len*sizeof(int))

        if self.approximate:
            mask_len = self._h_mask_approx(x_mask, y_mask, z_mask)
        else:
            mask_len = self._h_mask_exact(x_mask, y_mask, z_mask)

        for p from 0<=p<mask_len:
            x_temp = i + x_mask[p]
            y_temp = j + y_mask[p]
            z_temp = k + z_mask[p]

            if x_temp >= 0 and y_temp >= 0 and z_temp >= 0:
                cell = self.current_hash.get(x_temp, y_temp, z_temp)

                if cell == NULL:
                    continue

                h_local = self.radius_scale*fmax(cell.h_max, h)
                H = <int> ceil(h_local/self.h_sub)

                if abs(x_mask[p]) <= H and abs(y_mask[p]) <= H and abs(z_mask[p]) <= H:
                    x[length] = x_temp
                    y[length] = y_temp
                    z[length] = z_temp
                    length += 1

        free(x_mask)
        free(y_mask)
        free(z_mask)

        return length

    cpdef _refresh(self):
        cdef int i
        for i from 0<=i<self.narrays:
            del self.hashtable[i]
            self.hashtable[i] = new HashTable(self.table_size)
        self.current_hash = self.hashtable[self.src_index]

    @cython.cdivision(True)
    cpdef _bin(self, int pa_index, UIntArray indices):
        cdef NNPSParticleArrayWrapper pa_wrapper = self.pa_wrappers[pa_index]

        cdef double* src_x_ptr = pa_wrapper.x.data
        cdef double* src_y_ptr = pa_wrapper.y.data
        cdef double* src_z_ptr = pa_wrapper.z.data
        cdef double* src_h_ptr = pa_wrapper.h.data

        cdef int num_indices = indices.length

        cdef double* xmin = self.xmin.data
        cdef int c_x, c_y, c_z
        cdef unsigned int i
        cdef unsigned int idx

        self.h_sub = self.cell_size/self.H

        for i from 0<=i<num_indices:
            idx = indices.data[i]
            find_cell_id_raw(
                    src_x_ptr[idx] - xmin[0],
                    src_y_ptr[idx] - xmin[1],
                    src_z_ptr[idx] - xmin[2],
                    self.h_sub,
                    &c_x, &c_y, &c_z
                    )
            self._add_to_hashtable(pa_index, idx, src_h_ptr[idx], c_x, c_y, c_z)



""" Dumps VTK output files.

It takes a hdf or npz file as an input and output vtu file.
"""
from pysph import has_tvtk, has_pyvisfile
from pysph.solver.output import Output, load
from pysph.solver.utils import get_files

import numpy as np
import argparse
import sys
import os


class VTKOutput(Output):

    def __init__(self, scalars=None, **vectors):
        self.set_output_scalar(scalars)
        self.set_output_vector(**vectors)
        super(VTKOutput, self).__init__(True)

    def set_output_vector(self, **vectors):
        """
        Set the vector to dump in VTK output

        Parameter
        ----------

        vectors:
            Vectors to dump
            Example V=['u', 'v', 'z']
        """

        self.vectors = {}
        for name, vector in vectors.items():
            assert (len(vector) is 3)
            self.vectors[name] = vector

    def set_output_scalar(self, scalars=None):
        """
        Set the scalars to dump in VTK output

        Parameter
        ---------
        scalar_array: list
            The set of properties to dump
        """
        self.scalars = scalars

    def _get_scalars(self, arrays):

        if self.scalars is None:
            properties = list(arrays.keys())
        else:
            properties = self.scalars

        scalars = []
        for prop_name in properties:
            scalars.append((prop_name, arrays[prop_name]))
        return scalars

    def _get_vectors(self, arrays):
        vectors = []
        for prop_name, prop_list in self.vectors.items():
            vec = np.array([arrays[prop_list[0]], arrays[prop_list[1]],
                            arrays[prop_list[2]]])
            data = (prop_name, vec)
            vectors.append(data)
        return vectors

    def _dump(self, filename):
        for ptype, pdata in self.all_array_data.items():
            self._setup_data(pdata)
            try:
                fname, seq = filename.rsplit('_', 1)
                self._dump_arrays(fname + '_' + ptype + '_' + seq)
            except ValueError:
                self._dump_arrays(filename + '_' + ptype)

    def _setup_data(self, arrays):
        self.numPoints = arrays['x'].size
        self.points = np.array([arrays['x'], arrays['y'],
                                arrays['z']])
        self.data = []
        self.data.extend(self._get_scalars(arrays))
        self.data.extend(self._get_vectors(arrays))


class PyVisFileOutput(VTKOutput):

    def _dump_arrays(self, filename):
        from pyvisfile.vtk import (UnstructuredGrid, DataArray,
                                   AppendedDataXMLGenerator, VTK_VERTEX)
        n = self.numPoints
        da = DataArray("points", self.points)
        grid = UnstructuredGrid((n, da), cells=np.arange(n),
                                cell_types=np.asarray([VTK_VERTEX] * n))
        for name, field in self.data:
            da = DataArray(name, field)
            grid.add_pointdata(da)
        with open(filename + '.vtu', "w") as f:
            AppendedDataXMLGenerator(None)(grid).write(f)


class TVTKOutput(VTKOutput):
    def _dump_arrays(self, filename):
        from tvtk.api import tvtk
        n = self.numPoints
        cells = np.arange(n)
        cells.shape = (n, 1)
        cell_type = tvtk.Vertex().cell_type
        ug = tvtk.UnstructuredGrid(points=self.points.transpose())
        ug.set_cells(cell_type, cells)
        from mayavi.core.dataset_manager import DatasetManager
        dsm = DatasetManager(dataset=ug)
        for name, field in self.data:
            dsm.add_array(field.transpose(), name)
            dsm.activate(name)
        from tvtk.api import write_data
        write_data(ug, filename)


def dump_vtk(filename, particles, scalars=None, **vectors):
    """
    Parameter
    ----------

    filename: str
        Filename to dump to

    particles: sequence(ParticleArray)
        Sequence if particles arrays to dump

    scalars: list
        list of scalars to dump.

    vectors:
        Vectors to dump
        Example V=['u', 'v', 'z']
    """

    if has_pyvisfile():
        output = PyVisFileOutput(scalars, **vectors)
    elif has_tvtk():
        output = TVTKOutput(scalars, **vectors)
    else:
        msg = 'TVTK and pyvisfile Not present'
        raise ImportError(msg)
    output.dump(filename, particles, {})


def run(options):
    for fname in options.inputfile:
        if os.path.isdir(fname):
            files = get_files(fname)
            options.inputfile.extend(files)
            continue
        data = load(fname)
        particles = []
        for ptype, pdata in data['arrays'].items():
            particles.append(pdata)
        filename = os.path.splitext(fname)[0]
        outdir = options.outdir
        if outdir is not None:
            if not os.path.exists(outdir):
                os.makedirs(outdir)
            filename = os.path.join(outdir, os.path.basename(filename))
        dump_vtk(filename, particles, scalars=options.scalars,
                 velocity=['u', 'v', 'w'])


def main(argv=None):
    if argv is None:
        argv = sys.argv[1:]

    parser = argparse.ArgumentParser(
        prog='dump_vtk', description=__doc__, add_help=False
    )

    parser.add_argument(
        "-h", "--help", action="store_true", default=False, dest="help",
        help="show this help message and exit"
    )

    parser.add_argument(
        "-s", "--scalars",  metavar="scalars", type=str, default=None,
        help="scalar variables to dump in VTK output, provide a " +
        "comma-separated list, for example: -s rho,p,m"
    )

    parser.add_argument(
        "-d", "--outdir",  metavar="outdir", type=str, default=None,
        help="Directory to output VTK files"
    )

    parser.add_argument(
        "inputfile",  type=str, nargs='+',
        help=" list of input files  or/and directories (hdf5 or npz format)"
    )

    if len(argv) > 0 and argv[0] in ['-h', '--help']:
        parser.print_help()
        sys.exit()

    options, extra = parser.parse_known_args(argv)
    if options.scalars is not None:
        options.scalars = options.scalars.split(',')
    run(options)


if __name__ == '__main__':
    main()



"""
SPHysics case8 - dambreak with obstacles (30 minutes)
"""

from pysph.base.kernels import CubicSpline
from pysph.solver.solver import Solver
from pysph.solver.application import Application

from pysph.sph.integrator_step import WCSPHStep
from pysph.sph.integrator_step import TwoStageRigidBodyStep
from pysph.base.nnps import DomainManager
from pysph.base.utils import get_particle_array, get_particle_array_rigid_body

import numpy as np
from pysph.sph.equation import Group, Equation
from pysph.sph.scheme import AdamiHuAdamsScheme, WCSPHScheme, SchemeChooser
from pysph.sph.wc.edac import EDACScheme
from pysph.tools.geometry import remove_overlap_particles
from pysph.tools.geometry import get_2d_block, get_2d_wall
from pysph.sph.rigid_body import (BodyForce, RigidBodyCollision,
                                  RigidBodyMoments, RigidBodyMotion,
                                  RK2StepRigidBody, LiuFluidForce)
from pysph.examples.sphysics.periodic_rigidbody import GroupParticles


def get_geometry(dx_s=0.03, dx_f=0.03, hdx=1.3, r_f=100.0, r_s=100.0,
                 wall_l=4.0, wall_h=2.0, fluid_l=1., fluid_h=2., cube_s=0.25):
    wall_y1 = np.arange(dx_s, wall_h, dx_s)
    wall_xlayer = np.ones_like(wall_y1) * 2.0
    wall_x1 = []
    wall_x2 = []
    num_layers = 3
    for i in range(num_layers):
        wall_x1.append(wall_xlayer + i * dx_s)
        wall_x2.append(wall_xlayer - i * dx_s + wall_l / 4.0)
    wall_x1, wall_x2 = np.ravel(wall_x1), np.ravel(wall_x2)
    wall_y1 = np.tile(wall_y1, num_layers)
    wall_y2 = wall_y1
    w_center = np.array([wall_l / 2.0, 0.0])
    wall_x3, wall_y3 = get_2d_wall(dx_s, w_center, wall_l, num_layers, False)
    w_center = np.array([2.5, wall_h + dx_s / 2.0])
    wall_x4, wall_y4 = get_2d_wall(dx_s, w_center, 1.0, num_layers)
    wall_x = np.concatenate([wall_x1, wall_x2, wall_x3, wall_x4])
    wall_y = np.concatenate([wall_y1, wall_y2, wall_y3, wall_y4])
    r1 = np.ones_like(wall_x) * r_s
    m1 = r1 * dx_s * dx_s
    h1 = np.ones_like(wall_x) * dx_s * hdx
    cs1 = np.zeros_like(wall_x)
    rad1 = np.ones_like(wall_x) * dx_s
    wall = get_particle_array(name='wall', x=wall_x, y=wall_y, h=h1, rho=r1,
                              m=m1, cs=cs1, rad_s=rad1)
    f_center = np.array([3.0 * wall_l / 8.0, wall_h / 2.0])
    x2, y2 = get_2d_block(dx_f, fluid_l, fluid_h, f_center)
    r2 = np.ones_like(x2) * r_f
    m2 = r2 * dx_f * dx_f
    h2 = np.ones_like(x2) * dx_f * hdx
    cs2 = np.zeros_like(x2)
    rad2 = np.ones_like(x2) * dx_f
    fluid = get_particle_array(name='fluid', x=x2, y=y2, h=h2, rho=r2, m=m2,
                               cs=cs2, rad_s=rad2)
    center1 = np.array([wall_l / 8.0 + cube_s / 2.0,
                        wall_h / 4.0 + cube_s / 2.0])
    cube1_x, cube1_y = get_2d_block(dx_s, cube_s, cube_s, center1)
    b1 = np.zeros_like(cube1_x, dtype=int)
    center2 = np.array([3.0 * wall_l / 4.0 + cube_s / 2.0 + 3.0 * dx_s,
                        wall_h + cube_s / 2.0 + (num_layers + 1) * dx_s])
    cube2_x, cube2_y = get_2d_block(dx_s, cube_s, cube_s, center2)
    b2 = np.ones_like(cube2_x, dtype=int)
    b = np.concatenate([b1, b2])
    x3 = np.concatenate([cube1_x, cube2_x])
    y3 = np.concatenate([cube1_y, cube2_y])
    r3 = np.ones_like(x3) * r_s * 0.5
    m3 = r3 * dx_s * dx_s
    h3 = np.ones_like(x3) * dx_s * hdx
    cs3 = np.zeros_like(x3)
    rad3 = np.ones_like(x3) * dx_s
    cube = get_particle_array_rigid_body(
        name='cube', x=x3, y=y3, h=h3, cs=cs3, rho=r3, m=m3, rad_s=rad3,
        body_id=b)
    remove_overlap_particles(fluid, wall, dx_s, 2)
    return fluid, wall, cube


class Dambreak2D(Application):

    def add_user_options(self, group):
        group.add_argument(
            "--hdx", action="store", type=float, dest="hdx", default=1.3,
            help="h/dx value used in SPH to change the smoothness")
        group.add_argument(
            "--dx", action="store", type=float, dest="dx", default=0.03,
            help="spacing between the particles")

    def consume_user_options(self):
        self.hdx = self.options.hdx
        self.dx = self.options.dx
        self.h0 = self.hdx * self.dx
        self.dt = 0.25 * self.h0 / co

    def create_domain(self):
        return DomainManager(xmin=0.0, xmax=4.0, periodic_in_x=True)

    def create_particles(self):
        fluid, wall, cube = get_geometry(0.5 * self.dx, self.dx, self.hdx)
        self.scheme.setup_properties([fluid, wall, cube], clean=False)
        for p in ['u0', 'v0', 'w0', 'x0', 'y0', 'z0']:
            wall.add_property(p)
        for p in ['fx', 'fy', 'fz', 'V', 'arho']:
            cube.add_property(p)
        particles = [fluid, wall, cube]
        return particles

    def create_scheme(self):
        wcsph = WCSPHScheme(['fluid'], ['wall', 'cube'], dim=2, rho0=ro,
                            h0=0.03, hdx=1.3, hg_correction=True, c0=co,
                            gy=-9.81, alpha=alp, gamma=gamma, update_h=True)
        edac = EDACScheme(['fluid'], ['wall', 'cube'], dim=2, rho0=ro, c0=co,
                          alpha=alp, nu=0.0, h=0.03, gy=-9.81, clamp_p=True)
        aha = AdamiHuAdamsScheme(['fluid'], ['wall', 'cube'], dim=2, rho0=ro,
                                 h0=0.03, gamma=1.0, alpha=alp, gy=-9.81,
                                 nu=0.0, c0=co)
        return SchemeChooser(default='wcsph', aha=aha, wcsph=wcsph, edac=edac)

    def configure_scheme(self):
        s = self.scheme
        scheme = self.options.scheme
        if scheme == 'wcsph':
            s.configure(h0=self.h0, hdx=self.hdx)
        elif scheme == 'edac':
            s.configure(h=self.h0)
        step = dict(cube=RK2StepRigidBody())
        s.configure_solver(kernel=CubicSpline(dim=2), dt=self.dt, tf=3.0,
                           adaptive_timestep=False, extra_steppers=step)

    def create_equations(self):
        eqns = self.scheme.get_equations()
        eqn1 = Group(equations=[
            BodyForce(dest='cube', sources=None, gy=-9.81),
            RigidBodyCollision(dest='cube', sources=['wall', 'cube'],
                               kn=1.0e5, en=0.8),
            LiuFluidForce(dest='fluid', sources=['cube'])], real=False)
        eqn2 = Group(equations=[
            GroupParticles('cube', xmin=0.0, xmax=4.0, periodic_in_x=True)],
            real=False)
        eqn3 = Group(equations=[
            RigidBodyMoments(dest='cube', sources=None)], real=False)
        eqn4 = Group(equations=[
            RigidBodyMotion(dest='cube', sources=None)], real=False)
        eqns.append(eqn1)
        eqns.append(eqn2)
        eqns.append(eqn3)
        eqns.append(eqn4)
        return eqns


if __name__ == '__main__':
    l_dam = 4.0
    h_dam = 4.0
    h_fluid = 2.0
    l_fluid = 1.0
    gamma = 7.0
    alp = 0.2
    ro = 100.0
    co = 10.0 * np.sqrt(2.0 * 9.81 * h_fluid)
    app = Dambreak2D()
    app.run()



"""Simulate the hydrostatic box problem (30 minutes)

high density square region inside a low density square medium, in pressure
equilibrium, the solution should not evolve in time
"""

from pysph.sph.wc.crksph import CRKSPHScheme
from pysph.sph.scheme import (
    ADKEScheme, SchemeChooser, GasDScheme, GSPHScheme
)
from pysph.sph.gas_dynamics.psph import PSPHScheme
from pysph.sph.gas_dynamics.tsph import TSPHScheme
from pysph.base.utils import get_particle_array as gpa
from pysph.base.nnps import DomainManager
from pysph.solver.application import Application
from pysph.tools import uniform_distribution as ud
import numpy


class HydrostaticBox(Application):
    def initialize(self):
        self.xmin = 0.0
        self.xmax = 1.0
        self.ymin = 0.0
        self.ymax = 1.0
        self.gamma = 1.5
        self.p = 1
        self.rho0 = 1
        self.rhoi = 4
        self.nx = 50
        self.ny = self.nx
        self.dx = (self.xmax - self.xmin) / self.nx
        self.hdx = 1.5
        self.dt = 1e-3
        self.tf = 10

    def create_particles(self):
        data = ud.uniform_distribution_cubic2D(
            self.dx, self.xmin, self.xmax, self.ymin, self.ymax
        )
        x = data[0]
        y = data[1]
        box_indices = numpy.where(
            (x > 0.25) & (x < 0.75) & (y > 0.25) & (y < 0.75)
        )[0]
        rho = numpy.ones_like(x) * self.rho0
        rho[box_indices] = self.rhoi
        e = self.p / ((self.gamma - 1) * rho)
        m = self.dx * self.dx * rho
        h = self.hdx * self.dx

        fluid = gpa(
            name='fluid', x=x, y=y, p=self.p, rho=rho, e=e, u=0., v=0.,
            h=self.hdx*self.dx, m=m, h0=h
        )

        self.scheme.setup_properties([fluid])
        return [fluid]

    def create_domain(self):
        return DomainManager(
            xmin=self.xmin, xmax=self.xmax, ymin=self.ymin, ymax=self.ymax,
            periodic_in_x=True, periodic_in_y=True
        )

    def create_scheme(self):
        gsph = GSPHScheme(
            fluids=['fluid'], solids=[], dim=2,
            gamma=self.gamma, kernel_factor=1.0,
            g1=0., g2=0., rsolver=7, interpolation=1, monotonicity=1,
            interface_zero=True, hybrid=False, blend_alpha=5.0,
            niter=40, tol=1e-6, has_ghosts=True
        )

        mpm = GasDScheme(
            fluids=['fluid'], solids=[], dim=2, gamma=self.gamma,
            kernel_factor=1.2, alpha1=0, alpha2=0,
            beta=2.0, update_alpha1=False, update_alpha2=False,
            has_ghosts=True
        )

        crk = CRKSPHScheme(
            fluids=['fluid'], dim=2, rho0=0, c0=0, nu=0, h0=0, p0=0,
            gamma=self.gamma, cl=2, has_ghosts=True
        )
        adke = ADKEScheme(
            fluids=['fluid'], solids=[], dim=2, gamma=self.gamma,
            alpha=0.1, beta=0.1, k=1.5, eps=0., g1=0.1, g2=0.1,
            has_ghosts=True
        )
        psph = PSPHScheme(
            fluids=['fluid'], solids=[], dim=2, gamma=self.gamma,
            hfact=1.2
        )
        tsph = TSPHScheme(
            fluids=['fluid'], solids=[], dim=2, gamma=self.gamma,
            hfact=1.2
        )
        s = SchemeChooser(
            default='crksph', crksph=crk, adke=adke, mpm=mpm, gsph=gsph,
            psph=psph, tsph=tsph
        )
        return s

    def configure_scheme(self):
        s = self.scheme
        if self.options.scheme == 'gsph':
            s.configure_solver(
                dt=self.dt, tf=self.tf,
                adaptive_timestep=True, pfreq=50
            )
        elif self.options.scheme == 'mpm':
            s.configure(kernel_factor=1.2)
            s.configure_solver(dt=self.dt, tf=self.tf,
                               adaptive_timestep=True, pfreq=50)
        elif self.options.scheme == 'crksph':
            s.configure_solver(
                dt=self.dt, tf=self.tf, adaptive_timestep=False, pfreq=50
            )
        elif self.options.scheme == 'adke':
            s.configure_solver(dt=self.dt, tf=self.tf,
                               adaptive_timestep=False, pfreq=50)
        elif self.options.scheme in ['tsph', 'psph']:
            s.configure(hfact=1.2)
            s.configure_solver(dt=self.dt, tf=self.tf,
                               adaptive_timestep=False, pfreq=50)


if __name__ == "__main__":
    app = HydrostaticBox()
    app.run()



'''API for motion of points on the boundary of
the geometry'''

from pysph.sph.equation import Equation, Group
from pysph.sph.integrator_step import IntegratorStep
from pysph.sph.scheme import Scheme
from pysph.base.utils import get_particle_array
from pysph.tools.geometry import remove_overlap_particles
from itertools import combinations
from math import sqrt, sin
from compyle.api import declare
import numpy


def get_bounding_box(dx, x, y, z=[0], L=0.0, B=0.0, H=0.0):
    """Returns the bounding box required by the packing method
    """
    xmax, xmin, ymax, ymin, zmax, zmin = 0.0, 0.0, 0.0, 0.0, 0.0, 0.0
    if x is not None:
        xmax = max(x)
        xmin = min(x)
        ymax = max(y)
        ymin = min(y)
        zmax = max(z)
        zmin = min(z)

    if L < 1e-14:
        lenx = dx * int((xmax - xmin)/dx)
        leny = dx * int((ymax - ymin)/dx)
        lenz = dx * int((zmax - zmin)/dx)
    else:
        lenx = L
        leny = B
        lenz = H

    b0x = xmin - 0.5 * lenx
    b1x = xmax + 0.5 * lenx
    b0y = ymin - 0.5 * leny
    b1y = ymax + 0.5 * leny
    b0z = zmin - 0.5 * lenz
    b1z = zmax + 0.5 * lenz

    return b0x, b1x, b0y, b1y, b0z, b1z


def calculate_normal_2d_surface(boundary, shift, isclosed=True):
    """Calculate normals from the set of points arranged in a sequence

    Parameters
    ----------

    boundary: 2 dimensional numpy array of x and y coordinates
    shift: the signed shift value
    isclosed: True if the boundary is closed curve

    Returns
    -------

    xn: x coordinate of normal
    yn: y coordinate of normal
    x: shifted x coordinate of curve
    y: shifted y coordinate of curve
    area: length of the curve
    """
    import numpy as np
    xb = boundary[0]
    yb = boundary[1]

    xn = np.zeros(len(xb))
    yn = np.zeros(len(xb))
    xn0 = np.zeros(len(xb))
    yn0 = np.zeros(len(xb))
    xn1 = np.zeros(len(xb))
    yn1 = np.zeros(len(xb))
    x = np.zeros(len(xb))
    y = np.zeros(len(xb))
    area = np.zeros(len(xb))

    for i in range(len(xb)):
        next = i + 1
        prev = i - 1
        if isclosed:
            if i + 1 == len(xb):
                next = 0
            if i - 1 == -1:
                prev = len(xb) - 1
        else:
            if i + 1 == len(xb):
                next = i
            if i - 1 == -1:
                prev = 0
        dx0 = xb[next] - xb[i]
        dy0 = yb[next] - yb[i]
        d0 = sqrt(dx0**2 + dy0**2)
        xn0[i] = dy0 / (d0 + 1e-6)
        yn0[i] = -dx0 / (d0 + 1e-6)
        dx1 = xb[i] - xb[prev]
        dy1 = yb[i] - yb[prev]
        d1 = sqrt(dx1**2 + dy1**2)
        xn1[i] = dy1 / (d1 + 1e-6)
        yn1[i] = -dx1 / (d1 + 1e-6)
        area[i] = 0.5 * (d0 + d1)
        if not isclosed:
            if i == 0:
                xn1[i] = xn0[i]
                yn1[i] = yn0[i]
            elif i == len(xb) - 1:
                xn0[i] = xn1[i]
                yn0[i] = yn1[i]
        xn[i] = 0.5 * (xn0[i] + xn1[i])
        yn[i] = 0.5 * (yn0[i] + yn1[i])

    d = np.sqrt(xn**2 + yn**2)
    x = xb - shift * 0.5 * (xn0 + xn1) / d**2
    y = yb - shift * 0.5 * (yn0 + yn1) / d**2
    xn = xn / d
    yn = yn / d
    return xn, yn, x, y, area


def shift_surface_inside(x, y, shift, xn, yn,
                         z=numpy.array([0]),
                         zn=numpy.array([0])):
    """Shifted coodinated along xn, yn and zn

    Parameters
    ----------

    x: x coordinate of the boundary
    y: y coordinate of the boundary
    shift: the signed shift value
    xn: x coordinate of normal
    yn: y coordinate of normal
    z: z coordinated of the boundary
    zn: z coordinate of normal

    Returns
    -------

    x0: shifted x coordinate of boundary
    y0: shifted y coordinate of boundary
    z0: shifted z coordinate of boundary
    """
    x0 = x - shift * xn
    y0 = y - shift * yn
    z0 = z - shift * zn
    if len(z) == 1:
        return x0, y0
    else:
        return x0, y0, z0


def repair_boundary(x, y, hard):
    """Repair boundary near hard point

    Parameters
    ----------

    x: x coordinate of the boundary
    y: y coordinate of the boundary
    hard: list of hard points
    """
    N = len(x)
    for id in hard:
        id1 = id - 1
        id2 = id - 2
        id3 = id - 3
        id4 = id - 4
        if id1 < 0:
            id1 = N + id1
        if id2 < 0:
            id2 = N + id2
        if id3 < 0:
            id3 = N + id3
        if id4 < 0:
            id4 = N + id4
        dx = x[id] - x[id4]
        x[id3] = x[id] - 0.8 * dx
        x[id2] = x[id] - 0.5 * dx
        x[id1] = x[id] - 0.2 * dx
        dy = y[id] - y[id4]
        y[id3] = y[id] - 0.8 * dy
        y[id2] = y[id] - 0.5 * dy
        y[id1] = y[id] - 0.2 * dy

        d = (x[id] - x[id1])**2 + (y[id] - y[id1])**2
        id1 = id + 1
        id2 = id + 2
        id3 = id + 3
        id4 = id + 4
        if id1 >= N:
            id1 = id1 - N
        if id2 >= N:
            id2 = id2 - N
        if id3 >= N:
            id3 = id3 - N
        if id4 >= N:
            id4 = id4 - N
        dx = x[id4] - x[id]
        dy = y[id4] - y[id]
        d0 = dx**2 + dy**2
        fac = sqrt(d / d0)
        x[id3] = x[id] + 0.8 * dx
        x[id2] = x[id] + 0.5 * dx
        x[id1] = x[id] + fac * dx
        y[id3] = y[id] + 0.8 * dy
        y[id2] = y[id] + 0.5 * dy
        y[id1] = y[id] + fac * dy


def create_frozen_container_outer(
        dx, hdx, rho, bound, layers=8, dim=2, name='frozen'):
    """function to create frozen particle lattice of given
    size and spacing

    Parameters
    ----------

    dx: required resolution
    hdx: h/dx ratio
    rho: density of the particles
    L: length of the domain
    B: Breadth of the domain
    l: number of layers
    H: height of the domain
    dim: dimension of the problem
    name: name of the particle array

    Returns
    -------

    frozen: pysph.base.utils.ParticleArray class instance
    """
    import numpy as np
    eps = dx/10
    h = hdx * dx
    m = rho * dx**dim
    nl = layers * dx
    b = bound

    if dim == 2:
        x0, y0 = np.mgrid[b[0] + dx:b[1] - eps:2 * dx,
                          b[2]:b[3] + dx / 2 - eps:dx]
        x1, y1 = np.mgrid[b[0]:b[1] + dx - eps:2 * dx,
                          b[2] + dx / 2:b[3] - eps:dx]
        x0, y0 = [t.ravel() for t in (x0, y0)]
        x1, y1 = [t.ravel() for t in (x1, y1)]
        x = np.concatenate((x0, x1))
        y = np.concatenate((y0, y1))
        z = np.zeros_like(x)
    elif dim == 3:
        x0, y0, z0 = np.mgrid[b[0] + dx:b[1] - eps:2 * dx,
                              b[2]:b[3] - eps + dx / 2:dx,
                              b[4]:b[5] - eps + dx / 2:dx]
        x1, y1, z1 = np.mgrid[b[0]:b[1] - eps + dx:2 * dx,
                              b[2] + dx / 2:b[3] - eps:dx,
                              b[4] + dx / 2:b[5] - eps:dx]
        x0, y0, z0 = [t.ravel() for t in (x0, y0, z0)]
        x1, y1, z1 = [t.ravel() for t in (x1, y1, z1)]
        x = np.concatenate((x0, x1))
        y = np.concatenate((y0, y1))
        z = np.concatenate((z0, z1))

    inner = get_particle_array(x=x, y=y, z=z, m=m, rho=rho, h=h, name=name)

    if dim == 2:
        x0, y0 = np.mgrid[b[0] - 2 * nl:b[1] - eps + 2 * nl:dx,
                          b[2] - 2 * nl + dx / 2:b[3] - eps + 2 * nl:dx]
        print(x[0])
        x, y = [t.ravel() for t in (x0, y0)]
        z = np.zeros_like(x)
        cond = ~((x - (b[0] - eps - nl) > 1e-14) &
                 (x - (b[1] - eps + nl) < 1e-14) &
                 (y - (b[2] - eps - nl) > 1e-14) &
                 (y - (b[3] - eps + nl) < 1e-14))

    elif dim == 3:
        x0, y0, z0 = np.mgrid[b[0] - 2 * nl:b[1] - eps + 2 * nl:dx,
                              b[2] - 2 * nl + dx / 2:b[3] - eps + 2 * nl:dx,
                              b[4] - 2 * nl + dx / 2:b[5] - eps + 2 * nl:dx]
        x, y, z = [t.ravel() for t in (x0, y0, z0)]
        cond = ~((x - (b[0] - eps - nl) > 1e-14) &
                 (x - (b[1] + eps + nl) < 1e-14) &
                 (y - (b[2] - eps - nl) > 1e-14) &
                 (y - (b[3] + eps + nl) < 1e-14) &
                 (z - (b[4] - eps - nl) > 1e-14) &
                 (z - (b[5] + eps + nl) < 1e-14))

    frozen = get_particle_array(
        x=x[cond], y=y[cond], z=z[cond], m=m, rho=rho, h=h, name=name)
    ids = np.where(inner.h > -1)[0]
    print(ids)
    inner.extract_particles(ids, frozen)
    return frozen


def create_free_particles_outer(dx, hdx, rho, bound, layers=8, dim=2,
                                name='free'):
    """function to create free particle lattice of given size and spacing

    Parameters
    ----------

    dx: required resolution
    hdx: h/dx ratio
    rho: density of the particles
    L: length of the domain
    B: Breadth of the domain
    H: height of the domain
    dim: dimension of the problem
    name: name of the particle array

    Returns
    -------

    free: pysph.base.utils.ParticleArray class instance
    """
    import numpy as np
    eps = dx / 10
    h = hdx * dx
    m = rho * dx**dim
    nl = layers * dx
    b = bound
    if dim == 2:
        x0, y0 = np.mgrid[b[0] - nl + dx:b[1] - eps + nl:2 * dx,
                          b[2] - nl + dx / 2:b[3] - eps + nl:dx]
        x1, y1 = np.mgrid[b[0] - nl:b[1] - eps + nl:2 * dx,
                          b[2] - nl:b[3] - eps + nl:dx]
        x0, y0 = [t.ravel() for t in (x0, y0)]
        x1, y1 = [t.ravel() for t in (x1, y1)]
        x = np.concatenate((x0, x1))
        y = np.concatenate((y0, y1))
        z = np.zeros_like(x)
        print(x[0])
        cond = ~((x - b[0] + eps > 1e-14) & (x - b[1] - eps < 1e-14) &
                 (y - b[2] + eps > 1e-14) & (y - b[3] - eps < 1e-14))

    elif dim == 3:
        x0, y0, z0 = np.mgrid[b[0] - nl + dx:b[1] - eps + nl:2 * dx,
                              b[2] - nl + dx / 2:b[3] - eps + nl:dx,
                              b[4] - nl + dx / 2:b[5] - eps + nl:dx]
        x1, y1, z1 = np.mgrid[b[0] - nl:b[1] - eps + nl:2 * dx,
                              b[2] - nl:b[3] - eps + nl:dx,
                              b[4] - nl:b[5] - eps + nl:dx]
        x0, y0, z0 = [t.ravel() for t in (x0, y0, z0)]
        x1, y1, z1 = [t.ravel() for t in (x1, y1, z1)]
        x = np.concatenate((x0, x1))
        y = np.concatenate((y0, y1))
        z = np.concatenate((z0, z1))
        cond = ~((x - b[0] + eps > 1e-14) & (x - b[1] - eps < 1e-14) &
                 (y - b[2] + eps > 1e-14) & (y - b[3] - eps < 1e-14) &
                 (z - b[4] + eps > 1e-14) & (z - b[5] - eps < 1e-14))

    free = get_particle_array(
        x=x[cond], y=y[cond], z=z[cond], m=m,
        rho=rho, h=h, name=name)
    return free


def create_frozen_container_rect(
        dx, hdx, rho, bound, layers=3, dim=2, name='frozen'):
    """function to create frozen particle lattice of given
    size and spacing

    Parameters
    ----------

    dx: required resolution
    hdx: h/dx ratio
    rho: density of the particles
    L: length of the domain
    B: Breadth of the domain
    l: number of layers
    H: height of the domain
    dim: dimension of the problem
    name: name of the particle array

    Returns
    -------

    frozen: pysph.base.utils.ParticleArray class instance
    """
    import numpy as np
    eps = dx/10
    h = hdx * dx
    m = rho * dx**dim
    nl = layers * dx
    b = bound
    if dim == 2:
        x0, y0 = np.mgrid[b[0] - nl + dx / 2:b[1] - eps + nl:dx,
                          b[2] - nl + dx / 2:b[3] - eps + nl:dx]
        x, y = [t.ravel() for t in (x0, y0)]
        z = np.zeros_like(x)
        cond = ~((x - b[0] > 1e-14) & (x - b[1] < 1e-14) &
                 (y - b[2] > 1e-14) & (y - b[3] < 1e-14))

    elif dim == 3:
        x0, y0, z0 = np.mgrid[b[0] - nl + dx / 2:b[1] + nl:dx,
                              b[2] - nl + dx / 2:b[3] + nl:dx,
                              b[4] - nl + dx / 2:b[5] + nl:dx]
        x, y, z = [t.ravel() for t in (x0, y0, z0)]
        cond = ~((x - b[0] > 1e-14) & (x - b[1] < 1e-14) &
                 (y - b[2] > 1e-14) & (y - b[3] < 1e-14) &
                 (z - b[4] > 1e-14) & (z - b[5] < 1e-14))

    frozen = get_particle_array(
        x=x[cond], y=y[cond], z=z[cond], m=m,
        rho=rho, h=h, name=name)
    return frozen


def create_free_particles_rect(dx, hdx, rho, bound, dim=2, name='free'):
    """function to create free particle lattice of given size and spacing

    Parameters
    ----------

    dx: required resolution
    hdx: h/dx ratio
    rho: density of the particles
    L: length of the domain
    B: Breadth of the domain
    H: height of the domain
    dim: dimension of the problem
    name: name of the particle array

    Returns
    -------

    free: pysph.base.utils.ParticleArray class instance
    """
    import numpy as np
    eps = dx/10
    h = hdx * dx
    m = rho * dx**dim
    b = bound
    if dim == 2:
        x0, y0 = np.mgrid[b[0] + dx / 2:b[1]:dx,
                          b[2] + dx / 2:b[3]:dx]
        x, y = [t.ravel() for t in (x0, y0)]
        z = np.zeros_like(x)
    elif dim == 3:
        x0, y0, z0 = np.mgrid[b[0] + dx / 2:b[1]:dx,
                              b[2] + dx / 2:b[3]:dx,
                              b[4] + dx / 2:b[5]:dx]
        x, y, z = [t.ravel() for t in (x0, y0, z0)]
    free = get_particle_array(x=x, y=y, z=z, m=m, rho=rho, h=h, name=name)
    return free


def create_frozen_container(dx, hdx, rho, bound,
                            layers=3, dim=2,
                            name='frozen'):
    """function to create frozen particle lattice of given
    size and spacing

    Parameters
    ----------

    dx: required resolution
    hdx: h/dx ratio
    rho: density of the particles
    L: length of the domain
    B: Breadth of the domain
    layers: number of layers
    H: height of the domain
    dim: dimension of the problem
    name: name of the particle array

    Returns
    -------

    frozen: pysph.base.utils.ParticleArray class instance
    """
    import numpy as np
    eps = dx/10
    h = hdx * dx
    m = rho * dx**dim
    nl = layers * dx
    b = bound
    if dim == 2:
        x0, y0 = np.mgrid[b[0] - nl + dx:b[1] - eps + nl:2 * dx,
                          b[2] - nl + dx / 2:b[3] - eps + nl:dx]
        x1, y1 = np.mgrid[b[0] - nl:b[1] - eps + nl:2 * dx,
                          b[2] - nl:b[3] - eps + nl:dx]
        x0, y0 = [t.ravel() for t in (x0, y0)]
        x1, y1 = [t.ravel() for t in (x1, y1)]
        x = np.concatenate((x0, x1))
        y = np.concatenate((y0, y1))
        z = np.zeros_like(x)
        cond = ~((x - b[0] + eps > 1e-14) & (x - b[1] - eps < 1e-14) &
                 (y - b[2] + eps > 1e-14) & (y - b[3] - eps < 1e-14))

    elif dim == 3:
        x0, y0, z0 = np.mgrid[b[0] - nl + dx:b[1] - eps + nl:2 * dx,
                              b[2] - nl + dx / 2:b[3] - eps + nl:dx,
                              b[4] - nl + dx / 2:b[5] - eps + nl:dx]
        x1, y1, z1 = np.mgrid[b[0] - nl:b[1] - eps + nl:2 * dx,
                              b[2] - nl:b[3] - eps + nl:dx,
                              b[4] - nl:b[5] - eps + nl:dx]
        x0, y0, z0 = [t.ravel() for t in (x0, y0, z0)]
        x1, y1, z1 = [t.ravel() for t in (x1, y1, z1)]
        x = np.concatenate((x0, x1))
        y = np.concatenate((y0, y1))
        z = np.concatenate((z0, z1))
        cond = ~((x - b[0] + eps > 1e-14) & (x - b[1] - eps < 1e-14) &
                 (y - b[2] + eps > 1e-14) & (y - b[3] - eps < 1e-14) &
                 (z - b[4] + eps > 1e-14) & (z - b[5] - eps < 1e-14))

    frozen = get_particle_array(
        x=x[cond], y=y[cond], z=z[cond], m=m,
        rho=rho, h=h, name=name)
    return frozen


def create_free_particles(dx, hdx, rho, bound, dim=2, name='free'):
    """function to create free particle lattice of given size and spacing

    Parameters
    ----------

    dx: required resolution
    hdx: h/dx ratio
    rho: density of the particles
    L: length of the domain
    B: Breadth of the domain
    H: height of the domain
    dim: dimension of the problem
    name: name of the particle array

    Returns
    -------

    free: pysph.base.utils.ParticleArray class instance
    """
    import numpy as np
    eps = dx/10
    h = hdx * dx
    m = rho * dx**dim
    b = bound
    if dim == 2:
        x0, y0 = np.mgrid[b[0] + dx:b[1] - eps:2 * dx,
                          b[2]:b[3] - eps + dx / 2:dx]
        x1, y1 = np.mgrid[b[0]:b[1] - eps + dx:2 * dx,
                          b[2] + dx / 2:b[3] - eps:dx]
        x0, y0 = [t.ravel() for t in (x0, y0)]
        x1, y1 = [t.ravel() for t in (x1, y1)]
        x = np.concatenate((x0, x1))
        y = np.concatenate((y0, y1))
        z = np.zeros_like(x)
    elif dim == 3:
        x0, y0, z0 = np.mgrid[b[0] + dx:b[1] - eps:2 * dx,
                              b[2]:b[3] - eps + dx / 2:dx,
                              b[4]:b[5] - eps + dx / 2:dx]
        x1, y1, z1 = np.mgrid[b[0]:b[1] - eps + dx:2 * dx,
                              b[2] + dx / 2:b[3] - eps:dx,
                              b[4] + dx / 2:b[5] - eps:dx]
        x0, y0, z0 = [t.ravel() for t in (x0, y0, z0)]
        x1, y1, z1 = [t.ravel() for t in (x1, y1, z1)]
        x = np.concatenate((x0, x1))
        y = np.concatenate((y0, y1))
        z = np.concatenate((z0, z1))
    free = get_particle_array(x=x, y=y, z=z, m=m, rho=rho, h=h, name=name)
    return free


def create_surface_from_stl(
        filename, dx, hdx, rho, scale=1.0, shift=True, name='solid_nodes',
        hard={}, invert=False):
    """function to create solid nodes particle from a 3D Stl

    Parameters
    ----------

    filename: stl file name
    dx: required resolution
    hdx: h/dx ratio
    rho: density of the particles
    scale: scale factor
    shift: True if the geometry is to be shifted
    name: name of the particle array
    hard: dict of hard pint index with coordinates
    invert: True if normals to be inverted

    Returns
    -------

    solid_nodes: pysph.base.utils.ParticleArray class instance
    """
    import numpy as np
    from stl import stl
    from pysph.tools.geometry import evaluate_area_of_triangle
    import meshio
    data = meshio.read(filename, file_format="stl")
    triangles = data.cells_dict['triangle']
    normals = data.cell_data['facet_normals'][0]
    xc, yc, zc = [], [], []
    xn, yn, zn = normals[:, 0], normals[:, 1], normals[:, 2]
    area = []
    for tri in triangles:
        centroid = np.average(data.points[tri], axis=0)
        xc.append(centroid[0])
        yc.append(centroid[1])
        zc.append(centroid[2])
        _area = evaluate_area_of_triangle(data.points[tri])
        area.append(_area)

    xc, yc, zc, area = [np.array(t) for t in (xc, yc, zc, area)]

    h = hdx * dx
    m = dx * dx * dx * rho

    xx = scale * xc
    yx = scale * yc
    zx = scale * zc

    if invert:
        xn = -xn
        yn = -xn
        zn = -xn
    area = area * scale**2
    n_pnts = sum(area) / dx**2
    d = np.sqrt(xn**2 + yn**2 + zn**2)
    if shift:
        x, y, z = shift_surface_inside(
            x=xc, y=yc, z=zc, shift=dx / 2, xn=xn / d,
            yn=yn / d, zn=zn / d)
    else:
        x, y, z = xc.copy(), yc.copy(), zc.copy()
    solid_nodes = get_particle_array(
        x=x, y=y, z=z, m=m, rho=rho, h=h, name=name, xn=xn / d,
        yn=yn / d, zn=zn / d, area=area, hard=0.0)
    solid_nodes.add_property('xc')
    solid_nodes.add_property('yc')
    solid_nodes.add_property('zc')
    solid_nodes.add_constant('n_pnts', n_pnts)
    solid_nodes.xc[:] = xc
    solid_nodes.yc[:] = yc
    solid_nodes.zc[:] = zc
    for id in hard:
        solid_nodes.hard[id] = 1.0
    return solid_nodes


def create_surface_from_file(
        filename, points, dx, hdx, rho, isclosed, shift=True, invert=False,
        name='solid_nodes', hard={}):
    """function to create solid nodes particle from a x,y data file

    Parameters
    ----------

    filename: stl file name
    dx: required resolution
    hdx: h/dx ratio
    rho: density of the particles
    isclosed: True if the curve is closed
    shift: True if the geometry is to be shifted
    invert: True if normals to be inverted
    name: name of the particle array
    hard: dict of hard pint index with coordinates

    Returns
    -------

    solid_nodes: pysph.base.utils.ParticleArray class instance
    """
    import numpy as np
    boundary = None
    if filename is not None:
        xa, ya = np.loadtxt(filename, unpack=True)
        boundary = [xa, ya]
    else:
        boundary = points
        xa, ya = boundary[0], boundary[1]
    h = hdx * dx
    m = dx * dx * rho
    fact = 1.0
    _shift = 0.0
    if invert:
        fact = -1.0 * fact
    if shift:
        _shift = fact * dx / 2
    xn, yn, x, y, area = calculate_normal_2d_surface(
        boundary, _shift, isclosed=isclosed)
    repair_boundary(x, y, hard)
    n_pnts = sum(area) / dx
    solid_nodes = get_particle_array(
        x=x, y=y, z=0, m=m, rho=rho, h=h, xn=fact * xn,
        yn=fact * yn, zn=0, name=name, area=area, hard=0)
    solid_nodes.add_constant('n_pnts', n_pnts)
    for id in hard:
        solid_nodes.hard[id] = 1.0
    return solid_nodes


# intergrator
class InteriorStep(IntegratorStep):
    """Euler integrator for free particles
    """
    def stage1(self, d_idx, d_x, d_y, d_z,
               d_u, d_v, d_w, d_au, d_av, d_aw,
               dt=0.0):
        d_x[d_idx] = d_x[d_idx] + dt * d_u[d_idx]
        d_y[d_idx] = d_y[d_idx] + dt * d_v[d_idx]
        d_z[d_idx] = d_z[d_idx] + dt * d_w[d_idx]

        d_u[d_idx] = d_u[d_idx] + dt * d_au[d_idx]
        d_v[d_idx] = d_v[d_idx] + dt * d_av[d_idx]
        d_w[d_idx] = d_w[d_idx] + dt * d_aw[d_idx]


class SolidStep(IntegratorStep):
    """Euler integrator for boundary particles
    """
    def stage1(self, d_idx, d_x, d_y, d_z, d_u, d_v,
               d_w, d_au, d_av, d_aw, d_xr, d_yr,
               d_zr, d_hard, dt=0.0):
        if d_hard[d_idx] < 0.5:
            V = d_u[d_idx] * d_xr[d_idx] + d_v[d_idx] * d_yr[d_idx] + d_w[
                d_idx] * d_zr[d_idx]
            d_x[d_idx] = d_x[d_idx] + dt * d_xr[d_idx] * V
            d_y[d_idx] = d_y[d_idx] + dt * d_yr[d_idx] * V
            d_z[d_idx] = d_z[d_idx] + dt * d_zr[d_idx] * V

            d_u[d_idx] = d_u[d_idx] + dt * d_au[d_idx]
            d_v[d_idx] = d_v[d_idx] + dt * d_av[d_idx]
            d_w[d_idx] = d_w[d_idx] + dt * d_aw[d_idx]


# Equations
# used in post process
class FindExternalParticles(Equation):
    """Tag the neighboring particles inside and
     outside based on projection

    Parameters:

    dest: pysph.base.utils.ParticleArray
        Free particles array

    sources: list of pysph.base.utils.ParticleArray
        Boundary nodes array
    """
    def initialize(self, d_idx, d_interior):
        d_interior[d_idx] = 0

    def loop(self, d_idx, s_idx, s_xn, s_yn, s_zn, XIJ, d_neartag, d_interior,
             RIJ, d_h, s_hard):
        if d_neartag[d_idx] == s_idx:
            proj = XIJ[0] * s_xn[s_idx] + XIJ[1] * s_yn[s_idx] + XIJ[2] * s_zn[
                s_idx]
            if proj > 1e-14:
                d_interior[d_idx] = 1
            else:
                d_interior[d_idx] = -1


class FindNearNodes(Equation):
    """Tag the neighboring particles inside if the dest particle is interior

    Parameters:

    dest: pysph.base.utils.ParticleArray
        Free particles array

    sources: list of pysph.base.utils.ParticleArray
        Free particles array
    """
    def loop(self, d_idx, s_idx, d_interior, s_interior, d_neartag):
        if d_interior[d_idx] == 0:
            if s_interior[s_idx] == 1:
                d_interior[d_idx] = 1
            elif s_interior[s_idx] == -1:
                d_interior[d_idx] = -1


###########################################################
class SPHApprox(Equation):
    """Standard SPH approximation

    Parameters:

    dest: pysph.base.utils.ParticleArray
        any particles array

    sources: list of pysph.base.utils.ParticleArray
        any particles array other than dest
    """
    def loop(self, d_idx, d_f, s_m, s_rho, s_f, s_idx, WIJ):
        d_f[d_idx] += s_f[s_idx] * s_m[s_idx] * WIJ / s_rho[s_idx]


class SPHDerivativeApprox(Equation):
    """Standard SPH derivative approximation

    Parameters:

    dest: pysph.base.utils.ParticleArray
        any particles array

    sources: list of pysph.base.utils.ParticleArray
        any particles array other than dest
    """
    def loop(self, d_idx, d_df, s_m, s_rho, s_f, s_idx, DWIJ):
        d_df[d_idx] += s_f[s_idx] * s_m[s_idx] * DWIJ[0] / s_rho[s_idx]


class FindNearestNode(Equation):
    """Find nearest boundary node

    Parameters:

    dest: pysph.base.utils.ParticleArray
        Free or boundary particle

    sources: list of pysph.base.utils.ParticleArray
        Boundary nodes array
    """
    def __init__(self, dest, sources, fac=1.0):
        self.fac = fac
        super(FindNearestNode, self).__init__(dest, sources)

    def initialize(self, d_idx, d_nearest, d_neartag, d_xn, d_yn, d_zn,
                   d_hard):
        d_nearest[d_idx] = 10000.0
        d_neartag[d_idx] = -1
        if d_hard[d_idx] < 0.5:
            d_xn[d_idx] = 0.0
            d_yn[d_idx] = 0.0
            d_zn[d_idx] = 0.0

    def loop(self, d_idx, s_idx, RIJ, d_nearest, d_neartag, t, d_hard, d_h,
             d_xn, d_yn, d_zn, s_xn, s_yn, s_zn, s_hard):
        if (RIJ < d_nearest[d_idx]) and (d_hard[d_idx] < 0.5):
            if (s_hard[s_idx] < 0.5) and (RIJ - self.fac * d_h[d_idx] < 1e-14):
                d_nearest[d_idx] = RIJ
                d_neartag[d_idx] = s_idx
                d_xn[d_idx] = s_xn[s_idx]
                d_yn[d_idx] = s_yn[s_idx]
                d_zn[d_idx] = s_zn[s_idx]


class EvaluateAdaptiveTime(Equation):
    """Find new timestep

    Parameters:

    dest: pysph.base.utils.ParticleArray
        Free or boundary particle

    sources: None
    """
    def initialize(self, d_idx, d_dt_adapt, d_u, d_v, d_w, d_au, d_av, d_aw,
                   dt, t, d_h, d_pb, d_nu):
        if t < 1e-14:
            d_dt_adapt[d_idx] = 1e-7
        else:
            dt_vel = 10000
            dt_visc = 10000
            Vx = d_u[d_idx] + d_au[d_idx] * dt
            Vy = d_v[d_idx] + d_av[d_idx] * dt
            Vz = d_w[d_idx] + d_aw[d_idx] * dt
            V = sqrt(Vx**2 + Vy**2 + Vz**2)
            if V > 1e-14:
                dt_visc = sqrt(0.1 * d_h[d_idx] / d_nu[0] / V)
            dt_pb = 0.1 * d_h[d_idx] / sqrt(d_pb[0])
            dt_ = min(dt_vel, dt_visc, dt_pb)

            d_dt_adapt[d_idx] = dt_


class SummationDensity(Equation):
    """Standard summation density

    Parameters:

    dest: pysph.base.utils.ParticleArray
        Free or boundary particle

    sources: list of pysph.base.utils.ParticleArray
        All particle arrays (not nodes)
    """
    def initialize(self, d_idx, d_V, d_rho):
        d_V[d_idx] = 0.0
        d_rho[d_idx] = 0.0

    def loop(self, d_idx, s_idx, d_V, d_rho, d_m, s_m, WIJ):
        d_V[d_idx] += WIJ
        d_rho[d_idx] += s_m[s_idx] * WIJ


class NumberDensityGradient(Equation):
    """Evaluate the number density gradient

    Parameters:

    dest: pysph.base.utils.ParticleArray
        Free or boundary particle

    sources: list of pysph.base.utils.ParticleArray
        All particle arrays (not nodes)
    """
    def initialize(self, d_idx, d_au, d_av, d_aw):
        d_au[d_idx] = 0.0
        d_av[d_idx] = 0.0
        d_aw[d_idx] = 0.0

    def loop(self, d_idx, s_idx, d_m, d_rho, s_rho, d_pb, d_au, d_av, d_aw,
             d_V, s_V, DWIJ, XIJ, s_m):

        Vi = d_m[d_idx] / d_rho[d_idx]
        Vj = s_m[s_idx] / s_rho[s_idx]

        tmp = -d_pb[0] * Vi * Vj / d_m[d_idx]

        d_au[d_idx] += tmp * DWIJ[0]
        d_av[d_idx] += tmp * DWIJ[1]
        d_aw[d_idx] += tmp * DWIJ[2]


class ViscousDamping(Equation):
    """Evaluate acceleration due to damping

    Parameters:

    dest: pysph.base.utils.ParticleArray
        Free or boundary particle

    sources: list of pysph.base.utils.ParticleArray
        All particle arrays (not nodes)
    """
    def initialize(self, d_idx, d_au, d_av, d_aw):
        d_au[d_idx] = 0.0
        d_av[d_idx] = 0.0
        d_aw[d_idx] = 0.0

    def post_loop(self, d_idx, d_rho, d_m, d_V, d_au, d_av, d_aw, d_nu, d_u,
                  d_v, d_w, s_m, t):

        etai = d_nu[0]
        d_au[d_idx] += -etai * d_u[d_idx]
        d_av[d_idx] += -etai * d_v[d_idx]
        d_aw[d_idx] += -etai * d_w[d_idx]


class RepulsionForce(Equation):
    """accelearion due to repulsion force

    Parameters:

    dest: pysph.base.utils.ParticleArray
        free or boundary particles

    sources: pysph.base.utils.ParticleArray
        all particles (not nodes)

    hdx: double
        h/dx ratio
    """
    def __init__(self, dest, sources, hdx):
        self.hdx = hdx
        super(RepulsionForce, self).__init__(dest, sources)

    def initialize(self, d_idx, d_au, d_av, d_aw):
        d_au[d_idx] = 0.0
        d_av[d_idx] = 0.0
        d_aw[d_idx] = 0.0

    def loop(self, d_idx, d_au, d_av, d_aw, RIJ, XIJ, d_h, d_k):
        if RIJ > 1e-14:
            rij = RIJ
            xij = XIJ[0]
            yij = XIJ[1]
            zij = XIJ[2]
            if RIJ - d_h[d_idx] * 0.5 < 1e-14:
                rij = d_h[d_idx] * 0.5
                xij = xij / RIJ * rij
                yij = yij / RIJ * rij
                zij = zij / RIJ * rij
            c = self.hdx * d_h[d_idx] * 2. / 3.
            tmp = 12. * d_k[0] * (3 * c**2 / rij**5 - 2 * c / rij**4)
            if tmp > 1e-14:
                d_au[d_idx] += tmp * xij
                d_av[d_idx] += tmp * yij
                d_aw[d_idx] += tmp * zij


class FindBoundaryNodeDirection(Equation):
    """direction of motion of boundary particle

    Parameters:

    dest: pysph.base.utils.ParticleArray
        boundary particles

    sources: pysph.base.utils.ParticleArray
        corresponding boundary node array
    """
    def initialize(self, d_idx, d_dmin, d_xr, d_yr, d_zr):
        d_dmin[d_idx] = 10000
        d_xr[d_idx] = 0.0
        d_yr[d_idx] = 0.0
        d_zr[d_idx] = 0.0

    def loop(self, d_idx, d_dmin, XIJ, RIJ, VIJ, d_xr, d_yr, d_zr):
        dist = RIJ
        dir = VIJ[0] * XIJ[0] + VIJ[1] * XIJ[1] + VIJ[2] * XIJ[2]
        if dir < -1e-14:
            if d_dmin[d_idx] - dist > 1e-14:
                d_dmin[d_idx] = dist
                d_xr[d_idx] = -XIJ[0] / RIJ
                d_yr[d_idx] = -XIJ[1] / RIJ
                d_zr[d_idx] = -XIJ[2] / RIJ


class FilterLayers(Equation):
    """tag free particles near solid nodes

    Parameters:

    dest: pysph.base.utils.ParticleArray
        Boundary nodes

    sources: list of pysph.base.utils.ParticleArray
        all free particles
    """
    def loop(self, s_filter, s_idx, t):
        if t < 1e-14:
            s_filter[s_idx] = 1


class FindNearestNodeToHardPoint(Equation):
    """Find free particles to a geometry node

    Parameters:

    dest: pysph.base.utils.ParticleArray
        Boundary nodes

    sources: list of pysph.base.utils.ParticleArray
        all free particles
    """
    def initialize(self, d_idx, d_nearest, d_neartag, d_xn, d_yn, d_zn):
        d_nearest[d_idx] = 10000.0
        d_neartag[d_idx] = -1

    def loop(self, d_idx, s_idx, RIJ, d_nearest, d_neartag, t, d_hard, d_xn,
             d_yn, d_zn, s_xn, s_yn, s_zn, d_h):
        if (RIJ < d_nearest[d_idx]):
            d_nearest[d_idx] = RIJ
            d_neartag[d_idx] = s_idx


class ProjectionToSurfaceBoundary(Equation):
    """find the distance perpendicular to the surface for
    each free particle

    Parameters:

    dest: pysph.base.utils.ParticleArray
        free particles

    sources: list of pysph.base.utils.ParticleArray
        geometry nodes
    """
    def loop(self, d_idx, s_idx, XIJ, s_xn, s_yn, s_zn,
             d_neartag, d_u, d_v, d_w, d_h, d_nearest):
        projection = XIJ[0] * s_xn[s_idx] + XIJ[1] * s_yn[s_idx] + XIJ[
            2] * s_zn[s_idx]
        if (d_neartag[d_idx] == s_idx):
            d_nearest[d_idx] = projection


class ParticlePacking(Scheme):
    """An API for a hybrid particle packing scheme.

    Methods
    -------
    create_frozen_container(L, B, H=0, l=5, name='frozen')
        Creates the outer container of ghost particles

    create_free_particles(L, B, H=0, name='free')
        Creates array of free particles

    create_agitators(free, solid_nodes, L, B, H=0, name='agitator')
        Creates agitator particles [Not used]

    create_boundary_node(filename, scale=1.0, shift=True,
                         invert=False, name='solid_nodes', isclosed=True)
        Creates boundary node arrays

    _is_volume_converged(pa)
        Checks for convergence

    post_process(free, solid, solid_nodes, frozen, dx, filename)
        Method to split free particles into solid and fluid

    setup_hardpoints(pa_solid_nodes, pa_fluid, pa_solid)
        Sets a free particle near to a hard point on the boundary node

    _project_particles_to_boundary(pa_fluid, pa_solid)
        Method to project free particles to surface

    remove_duplicates(neartag, ids)
        Remove duplicates from the near tagged free particles list

    freeze_particles(pa_fluid, pa_frozen)
        freezes particles outside the filtered layers

    _check(particles, pa_fluid)
        Checks for convergence and decreases projection frequency

    post_step(particles, solver)
        methods runs post every iteration
    """
    def __init__(self, fluids, solids, frozen, dim, hdx=1.2,
                 dx=0.1, nu=None, pb=None, k=None, dfreq=-1,
                 hardpoints=None, use_prediction=None,
                 filter_layers=None, reduce_dfreq=None, tol=None):
        """Parameters
        ----------

        fluids: list
            list of free particles

        solids: dict
            dict with boundary and boundary nodes pair

        frozen: list
            list of frozen particles arrays

        dim: int
            dimension of the problem

        hdx: double
            h/dx ratio

        dx: double
            expected resolution

        nu: double
            damping constant

        pb: double
            background pressure

        k: double
            repulsion force constant

        dfreq: int
            projection frequency

        rho0: double
            expected density

        hardpoints: dict
            dictionary of hard point index and correspinding normal

        use_prediction: bool
            if true use prediction to projection first set of particles

        filter_layers: bool
            if True filter layers near to the surface to be packed

        reduce_dfreq: bool
            if True the projection frequency is reduced when projection stops

        tol : float
            tolerance for the convergence criteria
        """
        import numpy as np
        self.fluids = fluids
        self.solids = solids
        self.frozen = frozen
        self.solver = None
        self.dx = dx
        self.hdx = hdx
        self.dim = dim
        self.nu = nu
        self.pb = pb
        self.k = k
        self.hardpoints = {} if hardpoints is None else hardpoints
        self.rho0 = 1.0
        self.cutoff = 0.95
        self.nu_max = None
        self.dfreq = dfreq
        self.surface_points = 0
        self.do_check = False
        self.use_prediction = True if use_prediction is None\
            else use_prediction
        self.filter_layers = True if filter_layers is None\
            else filter_layers
        self.reduce_dfreq = True if reduce_dfreq is None\
            else reduce_dfreq
        self.converge = []
        self.blank = []
        self.divs = 10
        self.tol = tol

    def add_user_options(self, group):
        from pysph.sph.scheme import add_bool_argument

        group.add_argument("--dfreq", action="store", type=int, dest="dfreq",
                           default=None, help="particle deletion frequency.")

        group.add_argument("--pb", action="store", type=float,
                           dest="pb", default=None, help="Background pressure")

        group.add_argument("--nu", action="store", type=float,
                           dest="nu", default=None,
                           help="Dynamic viscosity")

        group.add_argument("--k", action="store", type=float,
                           dest="k", default=None,
                           help="Spring Constant")

        group.add_argument("--dx", action="store", type=float,
                           dest="dx", default=None,
                           help="Set particle spacing value")

        group.add_argument("--tol", action="store", type=float,
                           dest="tol", default=None,
                           help="tolerance for convergence")

        add_bool_argument(group, 'use-prediction',
                          dest='use_prediction',
                          help='use predicted number of points',
                          default=None)

        add_bool_argument(group, 'filter-layers',
                          dest='filter_layers',
                          help='use layered arrangement for packing',
                          default=None)

        add_bool_argument(group, 'reduce-dfreq',
                          dest='reduce_dfreq',
                          help='reduce update frequency for faster conv',
                          default=None)

    def consume_user_options(self, options):
        vars = [
            'dfreq', 'pb', 'nu', 'k', 'dx', 'use_prediction',
            'filter_layers', 'reduce_dfreq', 'tol'
        ]

        data = dict((var, self._smart_getattr(options, var)) for var in vars)
        self.configure(**data)

        dx = self.dx
        if self.pb is None:
            self.pb = 1.0

        if self.nu is None:
            if self.dim == 2:
                self.nu = .2/dx
            elif self.dim == 3:
                self.nu = 0.5/dx

        if self.k is None:
            if self.dim == 2:
                self.k = 0.004 * dx
            elif self.dim == 3:
                self.k = 0.006 * dx

        print(self.pb, self.k, self.nu)

        if self.dfreq < 0:
            self.dfreq = 50

        if self.tol is None:
            self.tol = 1e-2

    def create_frozen_container(self,
                                bound,
                                layers=5,
                                name='frozen',
                                outer=False,
                                rect=False):
        if outer:
            return create_frozen_container_outer(self.dx,
                                                 self.hdx,
                                                 1.0,
                                                 bound,
                                                 layers=layers,
                                                 dim=self.dim,
                                                 name=name)
        elif rect:
            return create_frozen_container_rect(self.dx,
                                                self.hdx,
                                                1.0,
                                                bound,
                                                layers=layers,
                                                dim=self.dim,
                                                name=name)
        else:
            return create_frozen_container(self.dx,
                                           self.hdx,
                                           1.0,
                                           bound,
                                           layers=layers,
                                           dim=self.dim,
                                           name=name)

    def create_free_particles(self,
                              bound,
                              layers=5,
                              name='free',
                              outer=False,
                              rect=False):
        if outer:
            return create_free_particles_outer(self.dx,
                                               self.hdx,
                                               1.0,
                                               bound,
                                               layers=layers,
                                               dim=self.dim,
                                               name=name)
        elif rect:
            return create_free_particles_rect(self.dx,
                                              self.hdx,
                                              1.0,
                                              bound,
                                              dim=self.dim,
                                              name=name)
        else:
            return create_free_particles(self.dx,
                                         self.hdx,
                                         1.0,
                                         bound,
                                         dim=self.dim,
                                         name=name)

    def create_boundary_node(self,
                             filename,
                             points=None,
                             scale=1.0,
                             shift=True,
                             invert=False,
                             name='solid_nodes',
                             isclosed=True):
        if self.dim == 2:
            return create_surface_from_file(filename,
                                            points,
                                            self.dx,
                                            self.hdx,
                                            1.0,
                                            isclosed,
                                            invert=invert,
                                            shift=shift,
                                            name=name,
                                            hard=self.hardpoints)
        elif self.dim == 3:
            return create_surface_from_stl(filename,
                                           self.dx,
                                           self.hdx,
                                           1.0,
                                           scale=scale,
                                           shift=shift,
                                           name=name,
                                           hard=self.hardpoints,
                                           invert=invert)

    def configure_solver(self, kernel=None, integrator_cls=None,
                         extra_steppers=None, **kw):
        from pysph.base.kernels import QuinticSpline
        if kernel is None:
            kernel = QuinticSpline(dim=self.dim)

        steppers = {}
        if extra_steppers is not None:
            steppers.update(extra_steppers)

        from pysph.sph.integrator import EulerIntegrator

        cls = EulerIntegrator
        for name in self.fluids:
            if name not in steppers:
                steppers[name] = InteriorStep()

        for name in self.solids:
            if name not in steppers:
                steppers[name] = SolidStep()

        integrator = cls(**steppers)

        from pysph.solver.solver import Solver
        self.solver = Solver(
            dim=self.dim, integrator=integrator, kernel=kernel, n_damp=10,
            adaptive_timestep=True, pfreq=3000, tf=200, max_steps=40000, **kw)

    def get_equations(self):

        all = self.fluids + list(self.solids.keys()) + self.frozen
        equations = []

        g1 = []
        for name in self.fluids:
            g1.append(
                FindNearestNode(dest=name, sources=list(self.solids.values())))
        for name in self.solids:
            g1.append(FindNearestNode(dest=name, sources=[self.solids[name]]))
            if self.filter_layers:
                g1.append(
                    FilterLayers(dest=self.solids[name], sources=self.fluids))
            if self.hardpoints:
                g1.append(
                    FindNearestNodeToHardPoint(dest=self.solids[name],
                                               sources=self.fluids))
        if len(self.solids.keys()) > 0:
            equations.append(Group(equations=g1, real=False))

        g2 = []
        for name in self.solids:
            g2.append(
                ProjectionToSurfaceBoundary(dest=name,
                                            sources=[self.solids[name]]))
        for name in self.fluids:
            if len(self.solids.keys()) > 0:
                g2.append(
                    ProjectionToSurfaceBoundary(dest=name,
                                                sources=list(
                                                    self.solids.values())))
        for name in self.solids:
            g2.append(
                FindBoundaryNodeDirection(dest=name,
                                          sources=[self.solids[name]]))
        for name in all:
            g2.append(SummationDensity(dest=name, sources=all))
        equations.append(Group(equations=g2, real=False))

        g3 = []
        dest = self.fluids + list(self.solids.keys())
        for name in dest:
            g3.append(NumberDensityGradient(dest=name, sources=all))
            if self.nu > 1e-14:
                g3.append(ViscousDamping(dest=name, sources=all))
        for name in self.fluids:
            g3.append(
                RepulsionForce(dest=name,
                               sources=all,
                               hdx=self.cutoff / self.hdx))
        source = self.fluids + self.frozen
        for name in self.solids:
            g3.append(
                RepulsionForce(dest=name,
                               sources=source,
                               hdx=self.cutoff / self.hdx))
        for name in self.solids:
            g3.append(
                RepulsionForce(dest=name,
                               sources=[name],
                               hdx=self.cutoff / self.hdx))
        equations.append(Group(equations=g3, real=False))

        g4 = []
        dest = self.fluids + list(self.solids.keys())
        for name in dest:
            g4.append(EvaluateAdaptiveTime(dest=name, sources=[]))
        equations.append(Group(equations=g4, real=False))

        print(equations)
        return equations

    def setup_properties(self, particles, clean=True):
        """Add following properties to the particle array

        x, y, z: coordinate of the position
        u, v, w: components of velocity
        p: pressure
        V: Volume
        h: support radius
        m: mass
        rho: density
        au, av, aw: components of acceleration
        dt_adapt: adaptive time step
        xr, yr, zr: unit direction of boundary particle motion
        x0, y0, z0: position at previous time step
        area: area of the boundary nodes
        hard: 1 if the particle is hard pint
        xc, yc, zc: centroid of boundary nodes
        xn, yn, zn: normal of boundary nodes
        dmin: minimum distance value
        nearest: distance of nearest particle
        neartag: id of nearest particle
        filter: 1 if the particle is near boundary node
        nu: damping constant
        pb: background pressure
        k: repulsion force constant
        """
        props = [
            'x', 'y', 'z', 'u', 'v', 'w', 'p', 'V', 'h', 'm', 'rho', 'au',
            'av', 'aw', 'dt_adapt', 'xr', 'yr', 'zr', 'x0', 'y0', 'z0',
            'area', 'hard', 'xc', 'yc', 'zc',
            'xn', 'yn', 'zn', 'dmin', 'nearest'
        ]
        output_props = [
            'x', 'y', 'z', 'u', 'v', 'w', 'V', 'rho', 'xn', 'yn', 'zn', 'm',
            'au', 'av', 'aw', 'h', 'xc', 'yc', 'zc', 'neartag', 'nearest',
            'filter', 'hard'
        ]

        newarr = []
        for pa in particles:
            prop_to_ensure = props.copy()
            self._ensure_properties(pa, prop_to_ensure, clean=False)
            pa.add_property('neartag', type='int')
            pa.add_property('filter', type='int')
            pa.add_constant('nu', self.nu)
            pa.add_constant('pb', self.pb)
            pa.add_constant('k', self.k)
            pa.set_output_arrays(output_props)
        particles.extend(newarr)

    def _is_volume_converged(self, pa):
        import numpy as np
        u = pa.u
        v = pa.v
        w = pa.w
        h = pa.h[0]
        vel = np.sqrt(u**2 + v**2 + w**2)
        maxvel = max(vel)
        rel_dist = maxvel * self.solver.dt / h * 100
        self.converge.append([rel_dist, self.solver.t])
        print('\nConvergence = ', rel_dist)
        if (rel_dist - self.tol < 1e-14) and (len(self.converge) > 10):
            self.solver.tf = self.solver.t

    def post_process(self, free, solid, solid_nodes, frozen, dx, filename):
        import numpy as np
        free_n = free.name

        if solid is None:
            import os
            xs, ys, zs, xf, yf, zf = None, None, None, None, None, None
            if os.path.exists(filename):
                data = np.load(filename)
                xs = data['xs']
                ys = data['ys']
                zs = data['zs']
                xf = data['xf']
                yf = data['yf']
                zf = data['zf']
                xf = np.concatenate((xf, free.x))
                yf = np.concatenate((yf, free.y))
                zf = np.concatenate((zf, free.z))
            else:
                xf = free.x
                yf = free.y
                zf = free.z
                xs = []
                ys = []
                zs = []
            return np.savez(filename, xs=xs, ys=ys, zs=zs, xf=xf, yf=yf, zf=zf)

        solid_nodes_n = solid_nodes.name
        frozen_n = frozen.name

        xb, yb, zb = None, None, None
        xi, yi, zi = None, None, None
        xf, yf, zf = None, None, None

        from pysph.tools.sph_evaluator import SPHEvaluator
        from pysph.sph.equation import Group

        arrays = []
        for pa in [free, solid_nodes, frozen]:
            arrays.append(pa)
            pa.add_property('interior', type='int')

        eqns = [
            Group(equations=[
                FindNearestNode(dest=free_n, sources=[solid_nodes_n], fac=4.0)
            ]),
            Group(equations=[
                FindExternalParticles(dest=free_n, sources=[solid_nodes_n])
            ]),
            Group(equations=[
                FindNearNodes(dest=free_n, sources=[free_n, frozen_n]),
                FindNearNodes(dest=frozen_n, sources=[free_n, frozen_n])
            ],
                  min_iterations=4, max_iterations=5, iterate=True)
        ]
        print(eqns)

        spheval = SPHEvaluator(
            arrays, eqns, kernel=self.solver.kernel,
            dim=self.dim, backend='cython')
        spheval.evaluate()

        isinterior = free.interior
        external = isinterior == 1
        internal = isinterior == -1
        xi = free.x[external]
        yi = free.y[external]
        zi = free.z[external]

        xf = free.x[internal]
        yf = free.y[internal]
        zf = free.z[internal]

        isinterior = frozen.interior
        filter = frozen.filter
        external = (isinterior == 1) & (filter < 0.5)
        internal = (isinterior == -1) & (filter < 0.5)
        f_xi = frozen.x[external]
        f_yi = frozen.y[external]
        f_zi = frozen.z[external]

        f_xf = frozen.x[internal]
        f_yf = frozen.y[internal]
        f_zf = frozen.z[internal]

        xb = solid.x
        yb = solid.y
        zb = solid.z

        xfluid = np.concatenate((xi, f_xi))
        yfluid = np.concatenate((yi, f_yi))
        zfluid = np.concatenate((zi, f_zi))

        xsolid = np.concatenate((xf, xb, f_xf))
        ysolid = np.concatenate((yf, yb, f_yf))
        zsolid = np.concatenate((zf, zb, f_zf))

        np.savez(filename, xs=xsolid, ys=ysolid, zs=zsolid,
                 xf=xfluid, yf=yfluid, zf=zfluid)

    def setup_hardpoints(self, pa_solid_nodes, pa_fluid, pa_solid):
        if self.solver.t < 1e-14:
            # replace nearest point from hard pint
            import numpy as np
            xh = pa_solid_nodes.x
            yh = pa_solid_nodes.y
            zh = pa_solid_nodes.z
            neartag = pa_solid_nodes.neartag
            h = pa_solid_nodes.h[0]
            m = pa_solid_nodes.m[0]
            rho = pa_solid_nodes.rho[0]

            x = []
            y = []
            z = []
            xn = []
            yn = []
            zn = []
            ids = []

            for id in list(self.hardpoints.keys()):
                x.append(xh[id])
                y.append(yh[id])
                z.append(zh[id])
                normal = self.hardpoints[id]
                xn.append(normal[0])
                yn.append(normal[1])
                zn.append(normal[2])
                ids.append(neartag[id])

            pa_fluid.remove_particles(ids)

            pa_add = get_particle_array(
                x=x, y=y, z=z, xn=xn, yn=yn, zn=zn, m=m, h=h, rho=rho,
                hard=1.0)
            pa_solid.add_particles(**pa_add.get_property_arrays())

    def _project_particles_to_boundary(self, pa_fluid, pa_solid):

        import numpy as np
        N = len(pa_solid.x)
        proj = pa_fluid.nearest
        neartag = pa_fluid.neartag
        h = pa_fluid.h[0]
        ids = None
        if ((self.solver.count == 0 and self.use_prediction)
                or (self.solver.count == self.dfreq and self.hardpoints
                    and self.use_prediction)):
            ids = np.argsort(abs(proj))
            n_pnts = int(0.90 * self.surface_points)
            ids = ids[0:n_pnts]
        else:
            num = self.divs
            if self.do_check:
                if len(self.blank) > 10:
                    if len(self.blank) == 11:
                        self.dfreq = 30
                    self.divs = 13
            for i in range(num):
                dist = 0.05 * float(i + 1) * h / self.hdx
                ids = np.where(abs(proj) - dist < 1e-14)[0]
                if len(ids) > 0.1 * self.surface_points:
                    ids = np.argsort(abs(proj))
                    num = int(0.1 * self.surface_points + 0.5)
                    ids = ids[0:num]
                if len(ids) > 0 and self.do_check and i > 9:
                    num = int(0.1 * (self.surface_points - N))
                    if num < 2:
                        num = 1
                    filter = 2
                    if num < len(ids):
                        filter = int(len(ids) / num)
                    ids = ids[0::filter]
                if len(ids) > 0:
                    break

        if len(ids) > 0:
            self.blank = []
        else:
            self.blank.append(0)
        uids = self.remove_duplicates(neartag[ids], ids)
        pa_fluid.extract_particles(uids, pa_solid)
        pa_fluid.remove_particles(uids)

        hard = pa_solid.hard
        cond = hard < 0.5
        proj = pa_solid.nearest
        xn = pa_solid.xn
        yn = pa_solid.yn
        zn = pa_solid.zn
        pa_solid.x[cond] -= proj[cond] * xn[cond]
        pa_solid.y[cond] -= proj[cond] * yn[cond]
        pa_solid.z[cond] -= proj[cond] * zn[cond]

    def remove_duplicates(self, neartag, ids):
        if len(ids) > 0:
            unique_ids = []
            new_tag = []
            for i in range(len(ids)):
                id = ids[i]
                tag = neartag[i]
                if tag not in new_tag:
                    new_tag.append(tag)
                    unique_ids.append(id)

            return unique_ids
        else:
            return ids

    def freeze_particles(self, pa_fluid, pa_frozen):
        import numpy as np

        # remove fluids
        filter = pa_fluid.filter
        ids = np.where(filter < 0.5)[0]
        pa_fluid.extract_particles(ids, pa_frozen)
        pa_fluid.remove_particles(ids)

        # remove it from adaptive time calc
        pa_frozen.dt_adapt[:] = np.inf

    def _check(self, particles, pa_fluid):
        import numpy as np
        if len(self.blank) > 11:
            self._is_volume_converged(pa_fluid)
            if self.dfreq > 2 and self.reduce_dfreq:
                self.dfreq = self.dfreq - 1

    def post_step(self, particles, solver):
        import numpy as np
        lookup = list(self.solids.values())
        pa_fluid = None
        pa_solid = None
        pa_frozen = None
        pa_solid_nodes = None
        dfreq = self.dfreq
        t = solver.t
        if solver.count % (dfreq) == 0:
            for name in self.solids:
                for pa in particles:
                    if name == pa.name:
                        pa_solid = pa
                    elif self.solids[name] == pa.name:
                        pa_solid_nodes = pa
                        self.surface_points = int(pa.n_pnts[0])
                        print('\npredicted surface pnts', self.surface_points)
                    elif self.fluids[0] == pa.name:
                        pa_fluid = pa
                    elif self.frozen[0] == pa.name:
                        pa_frozen = pa

                if t < 1e-14:
                    pa_frozen.filter[:] = 1  # done to remerge the fluid back
                if self.hardpoints:
                    self.setup_hardpoints(pa_solid_nodes, pa_fluid, pa_solid)
                if self.filter_layers and t < 1e-14:
                    self.freeze_particles(pa_fluid, pa_frozen)

                N = len(pa_solid.x)
                print('no of solids particles', N)
                if N > 1.5 * self.surface_points:
                    print(
                        'boundary attracting too many particles - increase'
                        'viscosity'
                    )
                    import sys
                    sys.exit()

                if self.hardpoints and solver.count == 0:
                    return
                self._project_particles_to_boundary(pa_fluid, pa_solid)

                if self.filter_layers:
                    self.do_check = True

                if len(self.blank) == 5:
                    self.dfreq = 5
                    self.do_check = True

                if (self.do_check):
                    self._check(particles, pa_fluid)

            if len(self.solids.keys()) == 0:
                for pa in particles:
                    if self.fluids[0] == pa.name:
                        pa_fluid = pa
                self._is_volume_converged(pa_fluid)



"""Helper module to easily create uniform distributions of particles"""

from __future__ import print_function
import numpy

def uniform_distribution_hcp2D(dx, xmin, xmax, ymin, ymax, adjust=False):
    """Hexagonal closed packing arrangement in 2D"""
    dy = 0.5 * numpy.sqrt(3.0) * dx
    dxb2 = 0.5 * dx
    dyb2 = 0.5 * dy
    
    # since we will be shifting each alternate row by dxb2, we use
    # xstart as dx/4
    xstart = xmin + 0.25 * dx

    ystart = ymin+dyb2

    # adjust ymax so that particles can fill a periodic region
    if adjust:
        _y = numpy.arange(ystart, ymax, dy)
        ymax = _y[-1] + 1.5*dy
    
    # create the points
    x, y = numpy.mgrid[xstart:xmax:dx,
                       ystart:ymax:dy]

    # each alternate row is shifted by dxb2
    x[:,::2] += dxb2
    x = x.ravel(); y = y.ravel()

    print('HCP packing domain: xmin, xmax, ymin, ymax =  ', xmin, xmax, ymin,
            ymax)
    print('HCP packing particles: xmin, xmax, ymin, ymax = ', x.min(), x.max(),
            y.min(), y.max())
    print('Particle spacings: dx, dy = ', dx, dy)
    print('Offset: xmin, xmax = ', x.min()-xmin, xmax-x.max())
    print('Offset: ymin, ymax = ', y.min()-ymin, ymax-y.max())

    return x, y, dx, dy, xmin, xmax, ymin, ymax

def uniform_distribution_cubic2D(dx, xmin, xmax, ymin, ymax, nrows=None):
    """Cubic lattice arrangement in 2D"""
    dy = dx
    dxb2 = 0.5 * dx
    dyb2 = 0.5 * dy
    
    if nrows is not None:
        ymax = nrows * dy

    xstart = xmin + dxb2
    ystart = ymin + dyb2
    x, y = numpy.mgrid[xstart:xmax:dx,
                       ystart:ymax:dy]

    x = x.ravel(); y = y.ravel()

    print('Cubic packing domain: xmin, xmax, ymin, ymax =  ', xmin, xmax, ymin,
            ymax)
    print('Cubic packing particles: xmin, xmax, ymin, ymax = ', x.min(),
            x.max(), y.min(), y.max())
    print('Particle spacings: dx, dy = ', dx, dy)
    print('Offset: xmin, xmax = ', x.min()-xmin, xmax-x.max())
    print('Offset: ymin, ymax = ', y.min()-ymin, ymax-y.max())

    return x, y, dx, dy, xmin, xmax, ymin, ymax

def get_number_density_hcp(dx, dy, kernel, h0):

    # create a dummy particle distribution with the reference spacings
    dxb2 = 0.5 * dx
    dyb2 = 0.5 * dy

    xstart = 0.25 * dx
    ystart = dyb2

    # create the points
    x, y = numpy.mgrid[xstart:1.0:dx,
                       ystart:1.0:dy]

    # each alternate row is shifted by dxb2
    x[:,::2] += dxb2

    # the target point
    nrows, ncols = x.shape
    x0, y0 = x[nrows/2, ncols/2], y[nrows/2, ncols/2]

    x = x.ravel(); y = y.ravel()

    # now do a kernel sum
    wij_sum = 0.0
    for i in range(x.size):
        xij = x0 - x[i]
        yij = y0 - y[i]
        zij = 0.0
        
        rij = numpy.sqrt( xij**2 + yij**2 + zij**2 )
        wij_sum += kernel.kernel( [xij, yij, zij], rij, h0 )
                
    return wij_sum



import numpy as np

from pysph.base.kernels import CubicSpline
from pysph.base.particle_array import ParticleArray

from pysph.sph.equation import Equation
from pysph.sph.integrator_step import IntegratorStep

from pysph.sph.integrator import EulerIntegrator
from pysph.solver.application import Application
from pysph.solver.solver import Solver
from pysph.base.reduce_array import parallel_reduce_array, serial_reduce_array


class TotalMass(Equation):
    def reduce(self, dst, t, dt):
        m = serial_reduce_array(dst.m, op='sum')
        dst.total_mass[0] = parallel_reduce_array(m, op='sum')


class DummyStepper(IntegratorStep):
    def initialize(self):
        pass

    def stage1(self):
        pass


class MyApp(Application):
    def create_particles(self):
        x = np.linspace(0, 10, 10)
        m = np.ones_like(x)
        y = np.zeros_like(x)
        z = np.zeros_like(x)
        h = np.ones_like(x) * 0.2

        fluid = ParticleArray(name='fluid', x=x, y=y, z=z, m=m, h=h)
        fluid.add_constant('total_mass', 0.0)
        return [fluid]

    def create_solver(self):
        dim = 1
        # Create the kernel
        kernel = CubicSpline(dim=dim)

        # Create the integrator.
        integrator = EulerIntegrator(fluid=DummyStepper())

        solver = Solver(kernel=kernel, dim=dim, integrator=integrator)
        solver.set_time_step(0.1)
        solver.set_final_time(0.1)
        # There is no need to write any output as the test below
        # computes the total mass.
        solver.set_disable_output(True)
        return solver

    def create_equations(self):
        return [TotalMass(dest='fluid', sources=['fluid'])]


def main():
    # Create the application.
    app = MyApp()
    app.run()

    fluid = app.particles[0]
    err = fluid.total_mass[0] - 10.0
    assert abs(err) < 1e-16, "Error: %s" % err


if __name__ == '__main__':
    main()



"""A very simple example to help benchmark PySPH. (2 minutes)

The example creates a cube shaped block of water falling in free-space under
the influence of gravity while solving the incompressible, inviscid flow
equations.  Only 5 time steps are solved but with a million particles.  It is
easy to change the number of particles by simply passing the command line
argument --np to a desired number::

    $ pysph run cube --np 2e6

To check the performance of PySPH using OpenMP one could try the following::

    $ pysph run cube --disable-output

    $ pysph run cube --disable-output --openmp

"""

import numpy

from pysph.base.kernels import CubicSpline
from pysph.base.utils import get_particle_array_wcsph
from pysph.solver.application import Application
from pysph.sph.scheme import WCSPHScheme

rho0 = 1000.0


class Cube(Application):
    def add_user_options(self, group):
        group.add_argument(
            "--np", action="store", type=float, dest="np", default=int(1e5),
            help="Number of particles in the cube (1e5 by default)."
        )

    def consume_user_options(self):
        self.hdx = 1.5
        self.dx = 1.0/pow(self.options.np, 1.0/3.0)

    def configure_scheme(self):
        self.scheme.configure(h0=self.hdx*self.dx, hdx=self.hdx)
        kernel = CubicSpline(dim=3)
        dt = 1e-4
        tf = 5e-4
        self.scheme.configure_solver(kernel=kernel, tf=tf, dt=dt)

    def create_scheme(self):
        co = 10.0
        s = WCSPHScheme(
            ['fluid'], [], dim=3, rho0=rho0, c0=co,
            h0=0.1, hdx=1.5, gz=-9.81, gamma=7.0,
            alpha=0.5, beta=0.0
        )
        return s

    def create_particles(self):
        dx = self.dx
        hdx = self.hdx
        xmin, xmax = 0.0, 1.0
        ymin, ymax = 0.0, 1.0
        zmin, zmax = 0.0, 1.0
        x, y, z = numpy.mgrid[xmin:xmax:dx, ymin:ymax:dx, zmin:zmax:dx]
        x = x.ravel()
        y = y.ravel()
        z = z.ravel()

        # set up particle properties
        h0 = hdx * dx

        volume = dx**3
        m0 = rho0 * volume

        fluid = get_particle_array_wcsph(name='fluid', x=x, y=y, z=z)
        fluid.m[:] = m0
        fluid.h[:] = h0

        fluid.rho[:] = rho0
        #nnps = LinkedListNNPS(dim=3, particles=[fluid])
        #nnps.spatially_order_particles(0)

        print("Number of particles:", x.size)
        fluid.set_lb_props( list(fluid.properties.keys()) )

        if fluid.gpu:
            fluid.gpu.push()

        return [fluid]


if __name__ == '__main__':
    app = Cube()
    app.run()



"""Simulate a 1D blast wave problem (30 seconds).
"""
from pysph.examples.gas_dynamics.shocktube_setup import ShockTubeSetup
from pysph.sph.scheme import ADKEScheme, GasDScheme, GSPHScheme, SchemeChooser
from pysph.sph.gas_dynamics.psph import PSPHScheme
from pysph.sph.gas_dynamics.tsph import TSPHScheme

# Numerical constants
dim = 1
gamma = 1.4
gamma1 = gamma - 1.0

# solution parameters
dt = 1e-6
tf = 0.0075

# domain size and discretization parameters
xmin = -0.5
xmax = 0.5


class Blastwave(ShockTubeSetup):

    def initialize(self):
        self.xmin = -0.5
        self.xmax = 0.5
        self.x0 = 0.0
        self.rhol = 1.0
        self.rhor = 1.0
        self.pl = 1000.0
        self.pr = 0.01
        self.ul = 0.0
        self.ur = 0.0

    def add_user_options(self, group):
        group.add_argument(
            "--hdx", action="store", type=float,
            dest="hdx", default=1.5,
            help="Ratio h/dx."
        )
        group.add_argument(
            "--nl", action="store", type=float, dest="nl", default=200,
            help="Number of particles in left region"
        )

    def consume_user_options(self):
        self.nl = self.options.nl
        self.hdx = self.options.hdx
        ratio = self.rhor/self.rhol
        self.nr = ratio*self.nl
        self.dxl = 0.5/self.nl
        self.dxr = 0.5/self.nr
        self.h0 = self.hdx * self.dxr
        self.hdx = self.hdx

    def create_particles(self):
        return self.generate_particles(xmin=-0.5, xmax=0.5, dxl=self.dxl,
                                       dxr=self.dxr, m=self.dxl, pl=self.pl,
                                       pr=self.pr, h0=self.h0, bx=0.03,
                                       gamma1=gamma1)

    def create_scheme(self):
        self.dt = dt
        self.tf = tf
        adke = ADKEScheme(
            fluids=['fluid'], solids=['boundary'], dim=dim, gamma=gamma,
            alpha=1, beta=1, k=1.0, eps=0.5, g1=0.2, g2=0.4)

        # Fix mpm scheme first
        mpm = GasDScheme(
            fluids=['fluid'], solids=['boundary'], dim=dim, gamma=gamma,
            kernel_factor=1.2, alpha1=1.0, alpha2=0.1,
            beta=2.0, update_alpha1=True, update_alpha2=True
        )

        gsph = GSPHScheme(
            fluids=['fluid'], solids=['boundary'], dim=dim, gamma=gamma,
            kernel_factor=1.0,
            g1=0.2, g2=0.4, rsolver=2, interpolation=1, monotonicity=1,
            interface_zero=True, hybrid=False, blend_alpha=2.0,
            niter=20, tol=1e-6
        )

        psph = PSPHScheme(
            fluids=['fluid'], solids=['boundary'], dim=dim, gamma=gamma,
            hfact=1.2
        )

        tsph = TSPHScheme(
            fluids=['fluid'], solids=['boundary'], dim=dim, gamma=gamma,
            hfact=1.2)

        s = SchemeChooser(default='adke', adke=adke, gsph=gsph,
                          psph=psph, tsph=tsph)
        return s


if __name__ == '__main__':
    app = Blastwave()
    app.run()
    app.post_process()
